import structlog

from py_semantic_taxonomy.cfg import get_settings
from py_semantic_taxonomy.dependencies import get_search_engine
from py_semantic_taxonomy.domain.entities import (
    Concept,
    SearchNotConfigured,
    SearchResult,
    UnknownLanguage,
)
from py_semantic_taxonomy.domain.hash_utils import hash_fnv64
from py_semantic_taxonomy.domain.ports import SearchEngine

logger = structlog.get_logger("py-semantic-taxonomy")


def c(language: str) -> str:
    return f"pyst-concepts-{language}"


class SearchService:
    def __init__(self, engine: SearchEngine | None = None):
        self.settings = get_settings()
        self.languages = {lang: c(lang) for lang in self.settings.languages}

        if self.settings.typesense_url == "missing" or not self.settings.typesense_url:
            self.configured = False
            logger.warning("Typesense not configured; search functionality won't work.")
            return

        self.engine = engine or get_search_engine()
        self.configured = True

        logger.info("Typesense URL: %s", self.settings.typesense_url)
        logger.info("Typesense languages: %s", ",".join(self.languages))
        logger.info("Typesense embedding model: %s", self.settings.typesense_embedding_model)

    def is_configured(self) -> bool:
        return self.configured

    async def initialize(self) -> None:
        if not self.is_configured():
            raise SearchNotConfigured

        await self.engine.initialize(list(self.languages.values()))

    async def reset(self) -> None:
        if not self.is_configured():
            raise SearchNotConfigured

        await self.engine.reset()

    async def create_concept(self, concept: Concept) -> None:
        if not self.is_configured():
            raise SearchNotConfigured

        for language, collection in self.languages.items():
            dct = concept.to_search_dict(language)
            if not (self.settings.typesense_exclude_if_missing_for_language) or dct["pref_label"]:
                await self.engine.create_concept(dct, collection)

    async def update_concept(self, concept: Concept) -> None:
        if not self.is_configured():
            raise SearchNotConfigured

        for language, collection in self.languages.items():
            dct = concept.to_search_dict(language)
            if not (self.settings.typesense_exclude_if_missing_for_language) or dct["pref_label"]:
                await self.engine.update_concept(dct, collection)

    async def delete_concept(self, iri: str) -> None:
        if not self.is_configured():
            raise SearchNotConfigured

        for collection in self.languages.values():
            await self.engine.delete_concept(hash_fnv64(iri), collection)

    async def search(
        self, query: str, language: str, semantic: bool = True, prefix: bool = False
    ) -> list[SearchResult]:
        if not self.is_configured():
            raise SearchNotConfigured

        if language not in self.languages:
            raise UnknownLanguage

        if prefix:
            semantic = False

        return await self.engine.search(
            query=query, collection=self.languages[language], semantic=semantic, prefix=prefix
        )

    async def suggest(self, query: str, language: str) -> list[SearchResult]:
        return await self.search(query=query, language=language, prefix=True)



CHANGE_NOTE = [
    {
        "http://purl.org/dc/terms/creator": [{"@id": "http://example.com/john.doe"}],
        "http://purl.org/dc/terms/issued": [
            {"@type": "http://www.w3.org/2001/XMLSchema#date", "@value": "1999-01-23"}
        ],
        "http://www.w3.org/1999/02/22-rdf-syntax-ns#value": [
            {"@language": "en", "@value": "Moved from under 'fruits' to under 'vegetables'"}
        ],
    },
]
HISTORY_NOTE = [
    {
        "http://purl.org/dc/terms/creator": [{"@id": "http://example.com/jane.doe"}],
        "http://purl.org/dc/terms/issued": [
            {"@type": "http://www.w3.org/2001/XMLSchema#date", "@value": "1971-01-23"}
        ],
        "http://www.w3.org/1999/02/22-rdf-syntax-ns#value": [
            {"@language": "en", "@value": "Previously thought we were talking about horseradish"}
        ],
    },
]
EDITORIAL_NOTE = [
    {
        "http://purl.org/dc/terms/creator": [{"@id": "http://example.com/jane.doe"}],
        "http://purl.org/dc/terms/issued": [
            {"@type": "http://www.w3.org/2001/XMLSchema#date", "@value": "2012-01-23"}
        ],
        "http://www.w3.org/1999/02/22-rdf-syntax-ns#value": [
            {"@language": "en", "@value": "Check if this includes species outside Animalia"}
        ],
    },
]
DEFINITION = [
    {
        "@language": "en",
        "@value": "Includes all sorts of live animals, including things you probably never heard of before, like psittaciformes. Those are birds like parrots, by the way. Also includes silly names like hinnies and asses, but not those kinds of asses, get your mind out of the gutter. We are very serious people doing very serious work with our asses.",
    }
]



import orjson
import structlog
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine

from py_semantic_taxonomy.adapters.persistence.tables import metadata_obj
from py_semantic_taxonomy.cfg import get_settings

logger = structlog.get_logger("py-semantic-taxonomy")


def create_engine(
    echo: bool = False,
) -> AsyncEngine:
    s = get_settings()
    if s.db_backend == "postgres":
        connection_str = (
            f"postgresql+asyncpg://{s.db_user}:{s.db_pass}@{s.db_host}:{s.db_port}/{s.db_name}"
        )
        logger.info("Using Postgres backend at %s:%s", s.db_host, s.db_port)
    elif s.db_backend == "sqlite":
        # Only for testing
        connection_str = "sqlite+aiosqlite:///:memory:"
        logger.info("Using in-memory SQLite backend")
    else:
        raise ValueError(f"Missing or incorrect database backend `PyST_db_backend`: {s.db_backend}")
    engine = create_async_engine(
        connection_str,
        json_serializer=lambda obj: orjson.dumps(obj).decode(),
        json_deserializer=lambda obj: orjson.loads(obj),
        echo=echo,
    )
    # Magic?
    # pool_pre_ping=True,
    # pool_size=global_settings.db_pool_size,
    # max_overflow=global_settings.db_max_overflow,
    return engine


async def init_db(engine: AsyncEngine) -> None:
    logger.info("Initializing relational database")
    async with engine.begin() as conn:
        await conn.run_sync(metadata_obj.create_all)


async def drop_db(engine: AsyncEngine) -> None:
    logger.info("Dropping relational database")
    async with engine.begin() as conn:
        await conn.run_sync(metadata_obj.drop_all)



from functools import lru_cache

from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    db_backend: str = "postgres"
    db_user: str = "missing"
    db_pass: str = "missing"
    db_host: str = "localhost"
    db_port: int = 5432
    db_name: str = "PyST"

    auth_token: str = "missing"

    typesense_url: str = "missing"
    typesense_api_key: str = "missing"
    typesense_embedding_model: str = "ts/all-MiniLM-L12-v2"
    typesense_exclude_if_missing_for_language: bool = True

    languages: list[str] = ["en", "de", "es", "fr", "pt", "it", "da"]

    # allow_origins: Set[str] = {
    #     "https://brightway.cauldron.ch",
    #     "https://brightway-lca.cloud",
    #     "https://brightway-lca.com",
    # }
    model_config = SettingsConfigDict(
        env_prefix="PyST_", env_file="pyst-config.env", env_file_encoding="utf-8"
    )


@lru_cache(maxsize=1)
def get_settings() -> Settings:
    return Settings()



from typing import Iterable
from urllib.parse import urlparse

import structlog
import typesense

from py_semantic_taxonomy.domain.entities import SearchResult

logger = structlog.get_logger("py-semantic-taxonomy")


def c(language: str) -> str:
    return f"pyst-concepts-{language}"


class TypesenseSearchEngine:
    def __init__(self, url: str, api_key: str, embedding_model: str):
        url = urlparse(url)
        self.client = typesense.Client(
            {
                "api_key": api_key,
                "nodes": [{"host": url.hostname, "port": url.port, "protocol": url.scheme}],
                "connection_timeout_seconds": 10,
            }
        )
        self.embedding_model = embedding_model

    async def _collection_labels(self) -> list[str]:
        collections = await self.client.collections.retrieve()
        return sorted([obj["name"] for obj in collections])

    async def initialize(self, collections: Iterable[str]) -> None:
        collection_labels = await self._collection_labels()
        logger.info("Existing typesense collections: %s", collection_labels)

        for name in collections:
            if name not in collection_labels:
                logger.info("Creating typesense collection %s", name)
                await self.client.collections.create(
                    {
                        "name": name,
                        "fields": [
                            {"name": "pref_label", "type": "string"},
                            {
                                "name": "pref_label_embedding",
                                "type": "float[]",
                                "embed": {
                                    "from": ["pref_label"],
                                    "model_config": {"model_name": self.embedding_model},
                                },
                            },
                            {"name": "alt_labels", "type": "string[]"},
                            {
                                "name": "alt_label_embedding",
                                "type": "float[]",
                                "embed": {
                                    "from": ["alt_labels"],
                                    "model_config": {"model_name": self.embedding_model},
                                },
                            },
                            {"name": "hidden_labels", "type": "string[]"},
                            {"name": "definition", "type": "string"},
                            {"name": "notation", "type": "string"},
                            {"name": "all_languages_pref_labels", "type": "string[]"},
                            {"name": "url", "type": "string"},
                        ],
                    }
                )

    async def reset(self) -> None:
        logger.warning("Resetting all Typesense collections")
        collection_labels = await self._collection_labels()
        for collection in collection_labels:
            await self.client.collections[collection].delete()

    # Can't use upsert because we have nested arrays:
    # https://github.com/typesense/typesense/issues/1043
    async def create_concept(self, concept: dict, collection: str) -> None:
        logger.debug("Creating concept %s in %s", concept["id"], collection)
        await self.client.collections[collection].documents.create(concept)

    async def update_concept(self, concept: dict, collection: str) -> None:
        logger.debug("Updating concept %s in %s", concept["id"], collection)
        await self.client.collections[collection].documents[concept.pop("id")].update(concept)

    async def delete_concept(self, id_: str, collection: str) -> None:
        logger.debug("Deleting concept %s in %s", id_, collection)
        await self.client.collections[collection].documents[id_].delete({"ignore_not_found": True})

    async def search(
        self, query: str, collection: str, semantic: bool, prefix: bool
    ) -> list[SearchResult]:
        without_semantic = (
            "pref_label,alt_labels,hidden_labels,notation,definition,all_languages_pref_labels"
        )
        with_semantic = "pref_label,pref_label_embedding,alt_labels,hidden_labels,notation,definition,all_languages_pref_labels"

        results = await self.client.collections[collection].documents.search(
            {
                "q": query,
                "query_by": with_semantic if semantic else without_semantic,
                "per_page": 50,
                "prefix": prefix,
                "exclude_fields": "pref_label_embedding",
            }
        )
        return SearchResult.from_typesense_results(results)



from pydantic import BaseModel, ConfigDict, Field

from py_semantic_taxonomy.adapters.routers.doc_examples import (
    CHANGE_NOTE,
    DEFINITION,
    EDITORIAL_NOTE,
    HISTORY_NOTE,
)
from py_semantic_taxonomy.domain.constants import RDF_MAPPING as RDF
from py_semantic_taxonomy.domain.constants import RelationshipVerbs as RV


class ErrorMessage(BaseModel):
    message: str
    detail: dict | None = None


class KOSCommon(BaseModel):
    id_: str = Field(
        alias=RDF["id_"],
        title="Object IRI (`@id`)",
        description="https://www.w3.org/TR/json-ld/#node-identifiers",
        example="http://data.europa.eu/xsp/cn2024/010021000090",
    )
    types: list[str] = Field(
        alias=RDF["types"],
        title="Object `@type`",
        description="https://www.w3.org/TR/json-ld/#specifying-the-type",
        example=["http://www.w3.org/2004/02/skos/core#Concept"],
    )
    pref_labels: list[dict[str, str]] = Field(
        alias=RDF["pref_labels"],
        title="SKOS preferred labels (one per language)",
        description="https://www.w3.org/TR/skos-primer/#secpref",
        example=[
            {"@value": "CHAPTER 1 - LIVE ANIMALS", "@language": "en"},
            {"@value": "CAP\u00cdTULO 1 - ANIMAIS VIVOS", "@language": "pt"},
        ],
    )
    status: list[dict[str, str]] = Field(
        alias=RDF["status"],
        title="BIBO status (accepted/draft/rejected)",
        description="https://github.com/dcmi/bibo/blob/main/rdf/bibo.ttl#L391",
        example=[{"@id": "http://purl.org/ontology/bibo/status/accepted"}],
    )
    definitions: list[dict[str, str]] = Field(
        alias=RDF["definitions"],
        default=[],
        title="SKOS definition (one per language)",
        description="https://www.w3.org/TR/skos-primer/#secdocumentation",
        example=DEFINITION,
    )
    notations: list[dict[str, str]] = Field(
        alias=RDF["notations"],
        default=[],
        description="https://www.w3.org/TR/skos-primer/#secnotations",
        title="SKOS notation (typed literal)",
        example=[
            {"@value": "01", "@type": "http://www.w3.org/1999/02/22-rdf-syntax-ns#PlainLiteral"}
        ],
    )
    change_notes: list[dict] = Field(
        alias=RDF["change_notes"],
        default=[],
        definition="https://www.w3.org/TR/skos-primer/#secdocumentation",
        title="SKOS change note with additional required fields",
        example=CHANGE_NOTE,
    )
    history_notes: list[dict] = Field(
        alias=RDF["history_notes"],
        default=[],
        definition="https://www.w3.org/TR/skos-primer/#secdocumentation",
        title="SKOS history note with additional required fields",
        example=HISTORY_NOTE,
    )
    editorial_notes: list[dict] = Field(
        alias=RDF["editorial_notes"],
        default=[],
        definition="https://www.w3.org/TR/skos-primer/#secdocumentation",
        title="SKOS editorial note with additional required fields",
        example=EDITORIAL_NOTE,
    )

    model_config = ConfigDict(extra="allow")

    def model_dump(self, exclude_unset=True, by_alias=True, *args, **kwargs):
        return super().model_dump(*args, exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)


class Concept(KOSCommon):
    schemes: list[dict] = Field(
        alias=RDF["schemes"],
        title="SKOS concept scheme",
        description="https://www.w3.org/TR/skos-primer/#secscheme",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/cn2024"}],
    )
    alt_labels: list[dict[str, str]] = Field(
        alias=RDF["alt_labels"],
        default=[],
        title="SKOS alternative labels. Can be more than one per language.",
        description="https://www.w3.org/TR/skos-primer/#secalt",
        examples=[
            {"@value": "Horsies, moo-moos, etc.", "@language": "en"},
        ],
    )
    hidden_labels: list[dict[str, str]] = Field(
        alias=RDF["hidden_labels"],
        default=[],
        title="SKOS hidden labels",
        description="https://www.w3.org/TR/skos-primer/#sechidden",
        example=[{"@language": "ine-pro", "@value": "ékwos"}],
    )
    top_concept_of: list[dict] = Field(
        alias=RDF["top_concept_of"],
        default=[],
        title="SKOS concept scheme if this concept is at top of hierarchy (maximum 1)",
        description="https://www.w3.org/TR/skos-primer/#secscheme",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/cn2024"}],
    )


class ConceptScheme(KOSCommon):
    created: list[dict] = Field(
        alias="http://purl.org/dc/terms/created",
        title="DCTerms created timestamp",
        description="https://www.dublincore.org/specifications/dublin-core/dcmi-terms/#http://purl.org/dc/terms/created",
        example=[
            {"@type": "http://www.w3.org/2001/XMLSchema#dateTime", "@value": "2023-10-11T13:59:56"}
        ],
    )
    creators: list[dict] = Field(
        alias="http://purl.org/dc/terms/creator",
        title="DCTerms creators list",
        description="https://www.dublincore.org/specifications/dublin-core/dcmi-terms/#http://purl.org/dc/elements/1.1/creator",
        example=[{"@id": "http://publications.europa.eu/resource/authority/corporate-body/ESTAT"}],
    )
    version: list[dict] = Field(
        alias="http://www.w3.org/2002/07/owl#versionInfo",
        title="OWL version info",
        description="https://www.w3.org/TR/owl-ref/#versionInfo-def",
        example=[{"@value": "2024"}],
    )


class Relationship(BaseModel):
    id_: str = Field(
        alias=RDF["id_"],
        title="Object IRI (`@id`)",
        description="https://www.w3.org/TR/json-ld/#node-identifiers",
        example="http://data.europa.eu/xsp/cn2024/010021000090",
    )
    broader: list[dict] = Field(
        alias=RV.broader,
        default=[],
        title="SKOS broader",
        description="https://www.w3.org/TR/skos-primer/#sechierarchy",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/010021000090"}],
    )
    narrower: list[dict] = Field(
        alias=RV.narrower,
        default=[],
        title="SKOS narrower",
        description="https://www.w3.org/TR/skos-primer/#sechierarchy",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/010100000080"}],
    )
    exact_match: list[dict] = Field(
        alias=RV.exact_match,
        default=[],
        title="SKOS exact match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010100000080"}],
    )
    close_match: list[dict] = Field(
        alias=RV.close_match,
        default=[],
        title="SKOS close match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010100000080"}],
    )
    broad_match: list[dict] = Field(
        alias=RV.broad_match,
        default=[],
        title="SKOS broad match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010021000090"}],
    )
    narrow_match: list[dict] = Field(
        alias=RV.narrow_match,
        default=[],
        title="SKOS narrow match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010100000080"}],
    )
    related_match: list[dict] = Field(
        alias=RV.related_match,
        default=[],
        title="SKOS related match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "https://www.wikidata.org/wiki/Q726"}],
    )

    model_config = ConfigDict(extra="forbid")

    def model_dump(self, exclude_unset=True, by_alias=True, *args, **kwargs):
        return super().model_dump(*args, exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)


class Correspondence(ConceptScheme):
    compares: list[dict] = Field(
        alias=RDF["compares"],
        title="List of `ConceptScheme` objects being compared",
        description="https://rdf-vocabulary.ddialliance.org/xkos.html#correspondences",
        example=[
            {"@id": "http://data.europa.eu/xsp/cn2023/cn2023"},
            {"@id": "http://data.europa.eu/xsp/cn2024/cn2024"},
        ],
    )
    made_ofs: list[dict] = Field(
        alias=RDF["made_ofs"],
        default=[],
        title="SKOS definition (one per language)",
        description="https://www.w3.org/TR/skos-primer/#secdocumentation",
        example=DEFINITION,
    )


class Association(BaseModel):
    id_: str = Field(
        alias=RDF["id_"],
        title="Object IRI (`@id`)",
        description="https://www.w3.org/TR/json-ld/#node-identifiers",
        example="http://data.europa.eu/xsp/cn2024/010021000090",
    )
    types: list[str] = Field(
        alias=RDF["types"],
        title="Object `@type`",
        description="https://www.w3.org/TR/json-ld/#specifying-the-type",
        example=["http://rdf-vocabulary.ddialliance.org/xkos#ConceptAssociation"],
    )
    source_concepts: list[dict] = Field(
        alias=RDF["source_concepts"],
        title="List of source `Concept` objects",
        description="https://rdf-vocabulary.ddialliance.org/xkos.html#correspondences",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010011000090"}],
    )
    target_concepts: list[dict] = Field(
        alias=RDF["target_concepts"],
        title="List of target `Concept` objects",
        description="https://rdf-vocabulary.ddialliance.org/xkos.html#correspondences",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/010011000090"}],
    )



from sqlalchemy import Table, cast, delete, func, insert, join, select, update
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine

from py_semantic_taxonomy.adapters.persistence.database import create_engine
from py_semantic_taxonomy.adapters.persistence.tables import (
    association_table,
    concept_scheme_table,
    concept_table,
    correspondence_table,
    relationship_table,
)
from py_semantic_taxonomy.domain.constants import (
    SKOS_HIERARCHICAL_RELATIONSHIP_PREDICATES,
    RelationshipVerbs,
)
from py_semantic_taxonomy.domain.entities import (
    Association,
    AssociationNotFoundError,
    Concept,
    ConceptNotFoundError,
    ConceptScheme,
    ConceptSchemeNotFoundError,
    Correspondence,
    CorrespondenceNotFoundError,
    DuplicateIRI,
    DuplicateRelationship,
    GraphObject,
    MadeOf,
    NotFoundError,
    Relationship,
)


class PostgresKOSGraphDatabase:
    def __init__(self, engine: AsyncEngine | None = None):
        self.engine = create_engine() if engine is None else engine

    async def get_object_type(self, iri: str) -> GraphObject:
        async with self.engine.connect() as conn:
            if await self._get_count_from_iri(conn, iri, concept_table):
                return Concept
            if await self._get_count_from_iri(conn, iri, concept_scheme_table):
                return ConceptScheme
            if await self._get_count_from_iri(conn, iri, correspondence_table):
                return Correspondence
            if await self._get_count_from_iri(conn, iri, association_table):
                return Association
        raise NotFoundError(f"Given IRI `{iri}` is not a known object")

    # Concepts

    async def _get_count_from_iri(self, connection: AsyncConnection, iri: str, table: Table) -> int:
        stmt = select(func.count("*")).where(table.c.id_ == iri)
        # TBD: This is ugly
        return (await connection.execute(stmt)).first()[0]

    async def concept_get(self, iri: str) -> Concept:
        async with self.engine.connect() as conn:
            stmt = select(concept_table).where(concept_table.c.id_ == iri)
            result = (await conn.execute(stmt)).first()
            if not result:
                raise ConceptNotFoundError
            await conn.rollback()
        return Concept(**result._mapping)

    async def concept_create(self, concept: Concept) -> Concept:
        async with self.engine.connect() as conn:
            count = await self._get_count_from_iri(conn, concept.id_, concept_table)
            if count:
                raise DuplicateIRI

            await conn.execute(
                insert(concept_table),
                [concept.to_db_dict()],
            )
            await conn.commit()
        return concept

    async def concept_update(self, concept: Concept) -> Concept:
        async with self.engine.connect() as conn:
            count = await self._get_count_from_iri(conn, concept.id_, concept_table)
            if not count:
                raise ConceptNotFoundError

            await conn.execute(
                update(concept_table)
                .where(concept_table.c.id_ == concept.id_)
                .values(**concept.to_db_dict())
            )
            await conn.commit()
        return concept

    async def concept_delete(self, iri: str) -> int:
        async with self.engine.connect() as conn:
            result = await conn.execute(delete(concept_table).where(concept_table.c.id_ == iri))
            await conn.commit()
        return result.rowcount

    async def concepts_get_for_scheme(
        self, concept_scheme_iri: str, top_concepts_only: bool
    ) -> list[Concept]:
        """Get all concepts that belong to a given concept scheme."""
        schemes = func.jsonb_array_elements(concept_table.c.schemes).table_valued(
            "value", joins_implicitly=True
        )
        top_concepts_of = func.jsonb_array_elements(concept_table.c.top_concept_of).table_valued(
            "value", joins_implicitly=True
        )

        async with self.engine.connect() as conn:
            # This approach seems to be faster (1 msec versus 4 msec; not sure this is correct or
            # significant) and feels more correct than using `@>` (contains); it would also allow
            # to get the concept scheme row which matches, though we already know its IRI so I'm
            # not sure how important that is. The previous approach was:
            # stmt = select(concept_table).where(
            #     concept_table.c.schemes.op("@>")([{"@id": concept_scheme_iri}])
            # )
            stmt = select(concept_table).where(
                schemes.c.value.op("->")("@id") == cast(concept_scheme_iri, JSONB)
            )
            if top_concepts_only:
                # Same as above; previous query:
                # stmt = stmt.where(
                #     concept_table.c.top_concept_of.op("@>")([{"@id": concept_scheme_iri}])
                # )
                stmt = stmt.where(
                    top_concepts_of.c.value.op("->")("@id") == cast(concept_scheme_iri, JSONB)
                )
            result = (await conn.execute(stmt.order_by(concept_table.c.id_))).fetchall()
            await conn.rollback()
        return [Concept(**row._mapping) for row in result]

    # ConceptScheme

    async def concept_scheme_get(self, iri: str) -> ConceptScheme:
        async with self.engine.connect() as conn:
            stmt = select(concept_scheme_table).where(concept_scheme_table.c.id_ == iri)
            result = (await conn.execute(stmt)).first()
            if not result:
                raise ConceptSchemeNotFoundError
            await conn.rollback()
        return ConceptScheme(**result._mapping)

    async def concept_scheme_list(self) -> list[ConceptScheme]:
        async with self.engine.connect() as conn:
            stmt = select(concept_scheme_table).order_by(concept_scheme_table.c.id_)
            result = (await conn.execute(stmt)).fetchall()
            await conn.rollback()
        return [ConceptScheme(**obj._mapping) for obj in result]

    async def concept_scheme_get_all_iris(self) -> list[str]:
        async with self.engine.connect() as conn:
            stmt = select(concept_scheme_table.c.id_)
            result = (await conn.execute(stmt)).scalars()
            await conn.rollback()
        return list(result)

    async def concept_scheme_create(self, concept_scheme: ConceptScheme) -> ConceptScheme:
        async with self.engine.connect() as conn:
            count = await self._get_count_from_iri(conn, concept_scheme.id_, concept_scheme_table)
            if count:
                raise DuplicateIRI

            await conn.execute(
                insert(concept_scheme_table),
                [concept_scheme.to_db_dict()],
            )
            await conn.commit()
        return concept_scheme

    async def concept_scheme_update(self, concept_scheme: ConceptScheme) -> ConceptScheme:
        async with self.engine.connect() as conn:
            count = await self._get_count_from_iri(conn, concept_scheme.id_, concept_scheme_table)
            if not count:
                raise ConceptSchemeNotFoundError

            await conn.execute(
                update(concept_scheme_table)
                .where(concept_scheme_table.c.id_ == concept_scheme.id_)
                .values(**concept_scheme.to_db_dict())
            )
            await conn.commit()
        return concept_scheme

    async def concept_scheme_delete(self, iri: str) -> int:
        async with self.engine.connect() as conn:
            result = await conn.execute(
                delete(concept_scheme_table).where(concept_scheme_table.c.id_ == iri)
            )
            await conn.commit()
        return result.rowcount

    async def known_concept_schemes_for_concept_hierarchical_relationships(
        self, iri: str
    ) -> list[str]:
        """Get list of all concept schemes for all known concepts with relationships to input iri"""
        h_verbs = [v for v in SKOS_HIERARCHICAL_RELATIONSHIP_PREDICATES if v in RelationshipVerbs]

        async with self.engine.connect() as conn:
            join_source = join(
                relationship_table,
                concept_table,
                relationship_table.c.source == concept_table.c.id_,
            )
            join_target = join(
                relationship_table,
                concept_table,
                relationship_table.c.target == concept_table.c.id_,
            )
            stmt = (
                select(concept_table.c.schemes)
                .select_from(join_source)
                .where(
                    relationship_table.c.target == iri,
                    relationship_table.c.predicate.in_(h_verbs),
                )
                .union(
                    select(concept_table.c.schemes)
                    .select_from(join_target)
                    .where(
                        relationship_table.c.source == iri,
                        relationship_table.c.predicate.in_(h_verbs),
                    )
                )
            )
            cursor = (await conn.execute(stmt)).scalars()
            results = {obj["@id"] for result in cursor for obj in result}
            await conn.rollback()
        return sorted(results)

    # Relationship

    async def relationships_get(
        self,
        iri: str,
        source: bool = True,
        target: bool = False,
        verb: RelationshipVerbs | None = None,
    ) -> list[Relationship]:
        if not source and not target:
            raise ValueError("Must choose at least one of source or target")
        async with self.engine.connect() as conn:
            rels = []
            if source:
                stmt = select(
                    # Exclude id field
                    relationship_table.c.source,
                    relationship_table.c.target,
                    relationship_table.c.predicate,
                ).where(relationship_table.c.source == iri)
                if verb is not None:
                    stmt = stmt.where(relationship_table.c.predicate == verb)
                result = await conn.execute(stmt)
                rels.extend([Relationship(**line._mapping) for line in result])
            if target:
                stmt = select(
                    relationship_table.c.source,
                    relationship_table.c.target,
                    relationship_table.c.predicate,
                ).where(relationship_table.c.target == iri)
                if verb is not None:
                    stmt = stmt.where(relationship_table.c.predicate == verb)
                result = await conn.execute(stmt)
                rels.extend([Relationship(**line._mapping) for line in result])
            await conn.rollback()
        return sorted(rels, key=lambda x: (x.source, x.target))

    async def relationships_create(self, relationships: list[Relationship]) -> list[Relationship]:
        async with self.engine.connect() as conn:
            try:
                await conn.execute(
                    insert(relationship_table), [obj.to_db_dict() for obj in relationships]
                )
            except IntegrityError as exc:
                await conn.rollback()
                err = exc._message()
                if (
                    # SQLite: Unit tests
                    "UNIQUE constraint failed: relationship.source, relationship.target"
                    in err
                ) or (
                    # Postgres: Integration tests
                    'duplicate key value violates unique constraint "relationship_source_target_uniqueness"'
                    in err
                ):
                    # Provide useful feedback by identifying which relationship already exists
                    for obj in relationships:
                        stmt = select(func.count("*")).where(
                            relationship_table.c.source == obj.source,
                            relationship_table.c.target == obj.target,
                        )
                        count = (await conn.execute(stmt)).first()[0]
                        if count:
                            raise DuplicateRelationship(
                                f"Relationship between source `{obj.source}` and target `{obj.target}` already exists"
                            )
                # Fallback - should never happen, but no one is perfect
                raise exc
            await conn.commit()
        return relationships

    async def relationships_delete(self, relationships: list[Relationship]) -> int:
        async with self.engine.connect() as conn:
            count = 0
            for rel in relationships:
                result = await conn.execute(
                    delete(relationship_table).where(
                        relationship_table.c.source == rel.source,
                        relationship_table.c.target == rel.target,
                        relationship_table.c.predicate == rel.predicate,
                    )
                )
                count += result.rowcount
            await conn.commit()
        return count

    async def relationship_source_target_share_known_concept_scheme(
        self, relationship: Relationship
    ) -> bool:
        async with self.engine.connect() as conn:
            stmt = select(concept_table.c.schemes).where(
                concept_table.c.id_.in_([relationship.source, relationship.target])
            )
            cursor = (await conn.execute(stmt)).scalars()
            results = [{obj["@id"] for obj in result} for result in cursor]
            await conn.rollback()
        return bool((len(results) < 2) or results[0].intersection(results[1]))

    # Correspondence

    async def correspondence_get(self, iri: str) -> Correspondence:
        async with self.engine.connect() as conn:
            stmt = select(correspondence_table).where(correspondence_table.c.id_ == iri)
            result = (await conn.execute(stmt)).first()
            if not result:
                raise CorrespondenceNotFoundError
            await conn.rollback()
        return Correspondence(**result._mapping)

    async def correspondence_create(self, correspondence: Correspondence) -> Correspondence:
        async with self.engine.connect() as conn:
            count = await self._get_count_from_iri(conn, correspondence.id_, correspondence_table)
            if count:
                raise DuplicateIRI

            await conn.execute(
                insert(correspondence_table),
                [correspondence.to_db_dict()],
            )
            await conn.commit()
        return correspondence

    async def correspondence_update(self, correspondence: Correspondence) -> Correspondence:
        async with self.engine.connect() as conn:
            count = await self._get_count_from_iri(conn, correspondence.id_, correspondence_table)
            if not count:
                raise CorrespondenceNotFoundError

            # Updates to `made_of` can only come via dedicated API calls
            values = correspondence.to_db_dict()
            if "made_ofs" in values:
                del values["made_ofs"]

            await conn.execute(
                update(correspondence_table)
                .where(correspondence_table.c.id_ == correspondence.id_)
                .values(**values)
            )
            await conn.commit()
        return correspondence

    async def correspondence_delete(self, iri: str) -> int:
        async with self.engine.connect() as conn:
            result = await conn.execute(
                delete(correspondence_table).where(correspondence_table.c.id_ == iri)
            )
            await conn.commit()
        return result.rowcount

    async def made_of_add(self, made_of: MadeOf) -> Correspondence:
        corr = await self.correspondence_get(iri=made_of.id_)
        existing = {assoc["@id"] for assoc in corr.made_ofs}
        new = [assoc for assoc in made_of.made_ofs if assoc["@id"] not in existing]
        async with self.engine.connect() as conn:
            await conn.execute(
                update(correspondence_table)
                .where(correspondence_table.c.id_ == made_of.id_)
                .values(made_ofs=sorted(corr.made_ofs + new, key=lambda x: x["@id"]))
            )
            await conn.commit()
        return await self.correspondence_get(iri=made_of.id_)

    async def made_of_remove(self, made_of: MadeOf) -> Correspondence:
        corr = await self.correspondence_get(iri=made_of.id_)
        to_remove = {assoc["@id"] for assoc in made_of.made_ofs}
        remaining = sorted(
            [assoc for assoc in corr.made_ofs if assoc["@id"] not in to_remove],
            key=lambda x: x["@id"],
        )
        async with self.engine.connect() as conn:
            await conn.execute(
                update(correspondence_table)
                .where(correspondence_table.c.id_ == made_of.id_)
                .values(made_ofs=remaining)
            )
            await conn.commit()
        return await self.correspondence_get(iri=made_of.id_)

    # Association

    async def association_get(self, iri: str) -> Association:
        async with self.engine.connect() as conn:
            stmt = select(association_table).where(association_table.c.id_ == iri)
            result = (await conn.execute(stmt)).first()
            if not result:
                raise AssociationNotFoundError
            await conn.rollback()
        return Association(**result._mapping)

    async def association_create(self, association: Association) -> Association:
        async with self.engine.connect() as conn:
            count = await self._get_count_from_iri(conn, association.id_, association_table)
            if count:
                raise DuplicateIRI

            await conn.execute(
                insert(association_table),
                [association.to_db_dict()],
            )
            await conn.commit()
        return association

    async def association_delete(self, iri: str) -> int:
        async with self.engine.connect() as conn:
            result = await conn.execute(
                delete(association_table).where(association_table.c.id_ == iri)
            )
            await conn.commit()
        return result.rowcount



#  Copyright 2014 Cilyan Olowen <gaknar@gmail.com>
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following disclaimer
#    in the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of the  nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# https://gist.github.com/Cilyan/9424144

import base64
import struct


def fnv64(data: bytes) -> int:
    hash_ = 0xCBF29CE484222325
    for b in data:
        hash_ *= 0x100000001B3
        hash_ &= 0xFFFFFFFFFFFFFFFF
        hash_ ^= b
    return hash_


def hash_fnv64(dn: str, salt: str = "abc123") -> str:
    # Turn dn into bytes with a salt, dn is expected to be ascii data
    data = salt.encode("ascii") + dn.encode("ascii")
    # Hash data
    hash_ = fnv64(data)
    # Pack hash (int) into bytes
    bhash = struct.pack("<Q", hash_)
    # Encode in base64. There is always a padding "=" at the end, because the
    # hash is always 64bits long. We don't need it.
    return base64.urlsafe_b64encode(bhash)[:-1].decode("ascii")



from fastapi import APIRouter, Request
from fastapi.responses import RedirectResponse

router = APIRouter(include_in_schema=False)


@router.get('/')
async def redirect_main_page(
    request: Request,
) -> RedirectResponse:
    return RedirectResponse(request.url_for('web_concept_schemes'))



from functools import lru_cache

from py_semantic_taxonomy.domain.ports import GraphService as GraphServiceProtocol
from py_semantic_taxonomy.domain.ports import KOSGraphDatabase, SearchEngine
from py_semantic_taxonomy.domain.ports import SearchService as SearchServiceProtocol


@lru_cache(maxsize=1)
def get_kos_graph() -> KOSGraphDatabase:
    # Import inside function so we don't import any database-related stuff in tests
    from py_semantic_taxonomy.adapters.persistence.graph import PostgresKOSGraphDatabase

    return PostgresKOSGraphDatabase()


@lru_cache(maxsize=1)
def get_search_engine() -> SearchEngine:
    from py_semantic_taxonomy.adapters.persistence.search_engine import TypesenseSearchEngine
    from py_semantic_taxonomy.cfg import get_settings

    settings = get_settings()
    return TypesenseSearchEngine(
        url=settings.typesense_url,
        api_key=settings.typesense_api_key,
        embedding_model=settings.typesense_embedding_model,
    )


@lru_cache(maxsize=1)
def get_graph_service() -> GraphServiceProtocol:
    from py_semantic_taxonomy.application.graph_service import GraphService

    return GraphService()


@lru_cache(maxsize=1)
def get_search_service() -> SearchServiceProtocol:
    from py_semantic_taxonomy.application.search_service import SearchService

    return SearchService()



import enum

BIBO = "http://purl.org/ontology/bibo/"
DCTERMS = "http://purl.org/dc/terms/"
OWL = "http://www.w3.org/2002/07/owl#"
SKOS = "http://www.w3.org/2004/02/skos/core#"
XKOS = "http://rdf-vocabulary.ddialliance.org/xkos#"

SKOS_ASSOCIATE_RELATIONSHIP_PREDICATES = {
    f"{SKOS}broadMatch",
    f"{SKOS}closeMatch",
    f"{SKOS}exactMatch",
    f"{SKOS}narrowMatch",
    f"{SKOS}relatedMatch",
}
SKOS_HIERARCHICAL_RELATIONSHIP_PREDICATES = {
    f"{SKOS}narrowerTransitive",
    f"{SKOS}narrower",
    f"{SKOS}broaderTransitive",
    f"{SKOS}broader",
}
SKOS_RELATIONSHIP_PREDICATES = SKOS_ASSOCIATE_RELATIONSHIP_PREDICATES.union(
    SKOS_HIERARCHICAL_RELATIONSHIP_PREDICATES
)


RDF_MAPPING: dict[str, str] = {
    "id_": "@id",
    "types": "@type",
    "alt_labels": f"{SKOS}altLabel",
    "change_notes": f"{SKOS}changeNote",
    "compares": f"{XKOS}compares",
    "created": f"{DCTERMS}created",
    "creators": f"{DCTERMS}creator",
    "definitions": f"{SKOS}definition",
    "editorial_notes": f"{SKOS}editorialNote",
    "hidden_labels": f"{SKOS}hiddenLabel",
    "history_notes": f"{SKOS}historyNote",
    "made_ofs": f"{XKOS}madeOf",
    "notations": f"{SKOS}notation",
    "pref_labels": f"{SKOS}prefLabel",
    "schemes": f"{SKOS}inScheme",
    "source_concepts": f"{XKOS}sourceConcept",
    "status": f"{BIBO}status",
    "target_concepts": f"{XKOS}targetConcept",
    "top_concept_of": f"{SKOS}topConceptOf",
    "version": f"{OWL}versionInfo",
}


class RelationshipVerbs(enum.StrEnum):
    broader = f"{SKOS}broader"
    narrower = f"{SKOS}narrower"
    exact_match = f"{SKOS}exactMatch"
    close_match = f"{SKOS}closeMatch"
    broad_match = f"{SKOS}broadMatch"
    narrow_match = f"{SKOS}narrowMatch"
    related_match = f"{SKOS}relatedMatch"


class AssociationKind(enum.StrEnum):
    simple = "simple"
    conditional = "conditional"



from typing import Protocol, runtime_checkable

from py_semantic_taxonomy.domain.constants import RelationshipVerbs
from py_semantic_taxonomy.domain.entities import (
    Association,
    Concept,
    ConceptScheme,
    Correspondence,
    GraphObject,
    MadeOf,
    Relationship,
    SearchResult,
)


@runtime_checkable
class KOSGraphDatabase(Protocol):
    async def get_object_type(self, iri: str) -> GraphObject: ...

    async def concept_get(self, iri: str) -> Concept: ...

    async def concept_create(self, concept: Concept) -> Concept: ...

    async def concept_update(self, concept: Concept) -> Concept: ...

    async def concept_delete(self, iri: str) -> int: ...

    async def concepts_get_for_scheme(
        self, concept_scheme_iri: str, top_concepts_only: bool
    ) -> list[Concept]: ...

    async def concept_scheme_get(self, iri: str) -> ConceptScheme: ...

    async def concept_scheme_get_all_iris(self) -> list[str]: ...

    async def concept_scheme_list(self) -> list[ConceptScheme]: ...

    async def concept_scheme_create(self, concept_scheme: ConceptScheme) -> ConceptScheme: ...

    async def concept_scheme_update(self, concept_scheme: ConceptScheme) -> ConceptScheme: ...

    async def concept_scheme_delete(self, iri: str) -> int: ...

    async def relationships_get(
        self, iri: str, source: bool, target: bool
    ) -> list[Relationship]: ...

    async def relationships_create(
        self, relationships: list[Relationship]
    ) -> list[Relationship]: ...

    async def relationships_delete(self, relationships: list[Relationship]) -> int: ...

    async def relationship_source_target_share_known_concept_scheme(
        self, relationship: Relationship
    ) -> bool: ...

    async def known_concept_schemes_for_concept_hierarchical_relationships(
        self, iri: str
    ) -> list[str]: ...

    async def correspondence_get(self, iri: str) -> Correspondence: ...

    async def correspondence_create(self, correspondence: Correspondence) -> Correspondence: ...

    async def correspondence_update(self, correspondence: Correspondence) -> Correspondence: ...

    async def correspondence_delete(self, iri: str) -> int: ...

    async def association_get(self, iri: str) -> Association: ...

    async def association_create(self, association: Association) -> Association: ...

    async def association_delete(self, iri: str) -> int: ...

    async def made_of_add(self, made_of: MadeOf) -> Correspondence: ...

    async def made_of_remove(self, made_of: MadeOf) -> Correspondence: ...


@runtime_checkable
class GraphService(Protocol):
    async def get_object_type(self, iri: str) -> GraphObject: ...

    async def concept_get(self, iri: str) -> Concept: ...

    async def concept_create(
        self, concept: Concept, relationships: list[Relationship] = []
    ) -> Concept: ...

    async def concept_update(self, concept: Concept) -> Concept: ...

    async def concept_delete(self, iri: str) -> None: ...

    async def concepts_get_for_scheme(
        self, concept_scheme_iri: str, top_concepts_only: bool
    ) -> list[Concept]: ...

    async def concept_scheme_get(self, iri: str) -> ConceptScheme: ...

    async def concept_scheme_get_all_iris(self) -> list[str]: ...

    async def concept_scheme_list(self) -> list[ConceptScheme]: ...

    async def concept_scheme_create(self, concept_scheme: ConceptScheme) -> ConceptScheme: ...

    async def concept_scheme_update(self, concept_scheme: ConceptScheme) -> ConceptScheme: ...

    async def concept_scheme_delete(self, iri: str) -> None: ...

    async def relationships_get(
        self,
        iri: str,
        source: bool = True,
        target: bool = False,
        verb: RelationshipVerbs | None = None,
    ) -> list[Relationship]: ...

    async def relationships_create(
        self, relationships: list[Relationship]
    ) -> list[Relationship]: ...

    async def relationships_delete(self, relationships: list[Relationship]) -> int: ...

    async def correspondence_get(self, iri: str) -> Correspondence: ...

    async def correspondence_create(self, correspondence: Correspondence) -> Correspondence: ...

    async def correspondence_update(self, correspondence: Correspondence) -> Correspondence: ...

    async def correspondence_delete(self, iri: str) -> None: ...

    async def made_of_add(self, made_of: MadeOf) -> Correspondence: ...

    async def made_of_remove(self, made_of: MadeOf) -> Correspondence: ...

    async def association_get(self, iri: str) -> Association: ...

    async def association_create(self, association: Association) -> Association: ...

    async def association_delete(self, iri: str) -> None: ...


@runtime_checkable
class SearchEngine(Protocol):
    async def initialize(self, collections: list[str]) -> None: ...

    async def reset(self) -> None: ...

    async def create_concept(self, concept: dict, collection: str) -> None: ...

    async def update_concept(self, concept: dict, collection: str) -> None: ...

    async def delete_concept(self, id_: str, collection: str) -> None: ...

    async def search(
        self, query: str, collection: str, semantic: bool, prefix: bool
    ) -> list[SearchResult]: ...


@runtime_checkable
class SearchService(Protocol):
    def is_configured(self) -> bool: ...

    async def initialize(self) -> None: ...

    async def reset(self) -> None: ...

    async def create_concept(self, concept: Concept) -> None: ...

    async def update_concept(self, concept: Concept) -> None: ...

    async def delete_concept(self, iri: str) -> None: ...

    async def search(
        self, query: str, language: str, semantic: bool = True, prefix: bool = False
    ) -> list[SearchResult]: ...

    async def suggest(self, query: str, language: str) -> list[SearchResult]: ...



from sqlalchemy import JSON, Column, Enum, Index, Integer, MetaData, String, Table, UniqueConstraint
from sqlalchemy.dialects.postgresql import JSONB

from py_semantic_taxonomy.domain.constants import AssociationKind, RelationshipVerbs

BetterJSON = JSON().with_variant(JSONB(), "postgresql")

metadata_obj = MetaData()


concept_table = Table(
    "concept",
    metadata_obj,
    Column("id_", String, primary_key=True),
    Column("types", BetterJSON, default=[]),
    Column("pref_labels", BetterJSON, default=[]),
    Column("schemes", BetterJSON, default=[]),
    Column("top_concept_of", BetterJSON, default=[]),
    Column("definitions", BetterJSON, default=[]),
    Column("notations", BetterJSON, default=[]),
    Column("alt_labels", BetterJSON, default=[]),
    Column("hidden_labels", BetterJSON, default=[]),
    Column("change_notes", BetterJSON, default=[]),
    Column("history_notes", BetterJSON, default=[]),
    Column("editorial_notes", BetterJSON, default=[]),
    Column("status", BetterJSON, default=[]),
    Column("extra", BetterJSON, default={}),
)

Index("concept_id_index", concept_table.c.id_)

concept_scheme_table = Table(
    "concept_scheme",
    metadata_obj,
    Column("id_", String, primary_key=True),
    Column("types", BetterJSON, default=[]),
    Column("pref_labels", BetterJSON, default=[]),
    Column("created", BetterJSON, default=[]),
    Column("creators", BetterJSON, default=[]),
    Column("version", BetterJSON, default=[]),
    Column("definitions", BetterJSON, default=[]),
    Column("notations", BetterJSON, default=[]),
    Column("change_notes", BetterJSON, default=[]),
    Column("history_notes", BetterJSON, default=[]),
    Column("editorial_notes", BetterJSON, default=[]),
    Column("status", BetterJSON, default=[]),
    Column("extra", BetterJSON, default={}),
)

Index("concept_scheme_id_index", concept_scheme_table.c.id_)

relationship_table = Table(
    "relationship",
    metadata_obj,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("source", String, nullable=False),
    Column("target", String, nullable=False),
    # https://docs.sqlalchemy.org/en/20/core/type_basics.html#sqlalchemy.types.Enum
    Column("predicate", Enum(RelationshipVerbs, values_callable=lambda x: [i.value for i in x])),
    UniqueConstraint("source", "target", name="relationship_source_target_uniqueness"),
)

Index("relationship_source_index", relationship_table.c.source)
Index("relationship_target_index", relationship_table.c.target)

correspondence_table = Table(
    "correspondence",
    metadata_obj,
    Column("id_", String, primary_key=True),
    Column("types", BetterJSON, default=[]),
    Column("compares", BetterJSON, default=[]),
    Column("made_ofs", BetterJSON, default=[]),
    Column("pref_labels", BetterJSON, default=[]),
    Column("created", BetterJSON, default=[]),
    Column("creators", BetterJSON, default=[]),
    Column("version", BetterJSON, default=[]),
    Column("definitions", BetterJSON, default=[]),
    Column("notations", BetterJSON, default=[]),
    Column("change_notes", BetterJSON, default=[]),
    Column("history_notes", BetterJSON, default=[]),
    Column("editorial_notes", BetterJSON, default=[]),
    Column("status", BetterJSON, default=[]),
    Column("extra", BetterJSON, default={}),
)

association_table = Table(
    "association",
    metadata_obj,
    Column("id_", String, primary_key=True),
    Column("types", BetterJSON, default=[]),
    Column("source_concepts", BetterJSON, default=[]),
    Column("target_concepts", BetterJSON, default=[]),
    Column("kind", Enum(AssociationKind, values_callable=lambda x: [i.value for i in x])),
    Column("extra", BetterJSON, default={}),
)



from enum import StrEnum
from typing import Annotated

from fastapi import APIRouter, Depends, Header, HTTPException, Request, status
from fastapi.responses import JSONResponse
from pydantic_settings import BaseSettings

import py_semantic_taxonomy.adapters.routers.request_dto as req
import py_semantic_taxonomy.adapters.routers.response_dto as response
from py_semantic_taxonomy.cfg import get_settings
from py_semantic_taxonomy.dependencies import get_graph_service, get_search_service
from py_semantic_taxonomy.domain import entities as de

router = APIRouter()


class Paths(StrEnum):
    concept = "/concept/"
    concept_scheme = "/concept_scheme/"
    relationship = "/relationships/"
    correspondence = "/correspondence/"
    association = "/association/"
    made_of = "/made_of/"
    search = "/concept/search/"
    suggest = "/concept/suggest/"


"""
In some cases, we will define a function input argument as a Pydantic class, but not use that
Pydantic class in the function code itself. This is because of the way we have defined our
Pydantic classes, which use child Pydantic models extensively, and our needs to use aliases. Aliases
are necessary because JSON-LD has keys like `@id` and
`http://www.w3.org/2004/02/skos/core#prefLabel`, which can't be used as Python identifiers.

When serializing data from Pydantic models, the behaviour of aliases is inconsistent. For example,
even if you specify `by_alias` true by default in child model `model_dump` functions, these nested
`model_dump` functions are actually not called when serializing the parent model, so `by_alias` is
not used.

The recommended solution is to use a custom `model_serializer`:

https://github.com/pydantic/pydantic/issues/8792

This approach would require duplicating `model_dump`, and would cause continuous frustration.

It's possible that this will change in the future:

https://github.com/pydantic/pydantic/issues/8379
"""


async def verify_auth_token(
    x_pyst_auth_token: Annotated[str, Header()] = "", settings: BaseSettings = Depends(get_settings)
):
    if x_pyst_auth_token != settings.auth_token:
        raise HTTPException(status_code=400, detail="X-PyST-Auth-Token header missing or invalid")


# Concept


@router.get(
    Paths.concept,
    summary="Get a `Concept` object",
    response_model=response.Concept,
    tags=["Concept"],
    responses={404: {"description": "Resource not found"}},
)
async def concept_get(
    iri: str,
    service=Depends(get_graph_service),
) -> response.Concept:
    try:
        obj = await service.concept_get(iri=iri)
        return response.Concept(**obj.to_json_ld())
    except de.ConceptNotFoundError:
        raise HTTPException(status_code=404, detail=f"Concept with IRI `{iri}` not found")


@router.post(
    Paths.concept,
    summary="Create a `Concept` object",
    response_model=response.Concept,
    dependencies=[Depends(verify_auth_token)],
    tags=["Concept"],
    responses={409: {"description": "Resource already exists"}},
)
async def concept_create(
    request: Request,
    concept: req.ConceptCreate,
    service=Depends(get_graph_service),
) -> response.Concept:
    try:
        incoming_data = await request.json()
        concept_obj = de.Concept.from_json_ld(incoming_data)
        relationships = de.Relationship.from_json_ld(incoming_data)
        result = await service.concept_create(concept=concept_obj, relationships=relationships)
        return response.Concept(**result.to_json_ld())
    except de.DuplicateIRI:
        raise HTTPException(
            status_code=409, detail=f"Concept with IRI `{concept_obj.id_}` already exists"
        )
    except (
        de.HierarchicRelationshipAcrossConceptScheme,
        de.DuplicateRelationship,
        de.ConceptSchemesNotInDatabase,
        de.HierarchyConflict,
    ) as err:
        raise HTTPException(status_code=422, detail=str(err))


@router.put(
    Paths.concept,
    summary="Update a `Concept` object",
    response_model=response.Concept,
    dependencies=[Depends(verify_auth_token)],
    tags=["Concept"],
    responses={404: {"description": "Resource not found"}},
)
async def concept_update(
    request: Request,
    concept: req.ConceptUpdate,
    service=Depends(get_graph_service),
) -> response.Concept:
    try:
        concept_obj = de.Concept.from_json_ld(await request.json())
        result = await service.concept_update(concept_obj)
        return response.Concept(**result.to_json_ld())
    except de.ConceptNotFoundError:
        raise HTTPException(
            status_code=404, detail=f"Concept with IRI `{concept_obj.id_}` not found"
        )
    except (
        de.RelationshipsInCurrentConceptScheme,
        de.ConceptSchemesNotInDatabase,
        de.HierarchyConflict,
    ) as err:
        raise HTTPException(status_code=422, detail=str(err))


@router.delete(
    Paths.concept,
    summary="Delete a `Concept` object",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(verify_auth_token)],
    tags=["Concept"],
    responses={404: {"description": "Resource not found"}},
)
async def concept_delete(
    iri: str,
    service=Depends(get_graph_service),
):
    try:
        await service.concept_delete(iri=iri)
    except de.ConceptNotFoundError as err:
        raise HTTPException(status_code=404, detail=str(err))


# Concept Scheme


@router.get(
    Paths.concept_scheme,
    summary="Get a `ConceptScheme` object or list all concept schemes",
    response_model=response.ConceptScheme | list[response.ConceptScheme],
    tags=["ConceptScheme"],
    responses={404: {"description": "Resource not found"}},
)
async def concept_scheme_get(
    iri: str | None = None,
    service=Depends(get_graph_service),
) -> response.ConceptScheme | list[response.ConceptScheme]:
    if iri is None:
        concept_schemes = await service.concept_scheme_list()
        return [response.ConceptScheme(**cs.to_json_ld()) for cs in concept_schemes]

    try:
        obj = await service.concept_scheme_get(iri=iri)
        return response.ConceptScheme(**obj.to_json_ld())
    except de.ConceptSchemeNotFoundError:
        raise HTTPException(status_code=404, detail=f"Concept Scheme with IRI `{iri}` not found")


@router.post(
    Paths.concept_scheme,
    summary="Create a `ConceptScheme` object",
    response_model=response.ConceptScheme,
    dependencies=[Depends(verify_auth_token)],
    tags=["ConceptScheme"],
)
async def concept_scheme_create(
    request: Request,
    concept_scheme: req.ConceptScheme,
    service=Depends(get_graph_service),
    responses={409: {"description": "Resource already exists"}},
) -> response.ConceptScheme:
    try:
        cs = de.ConceptScheme.from_json_ld(await request.json())
        result = await service.concept_scheme_create(cs)
        return response.ConceptScheme(**result.to_json_ld())
    except de.DuplicateIRI:
        raise HTTPException(
            status_code=409, detail=f"Concept Scheme with IRI `{cs.id_}` already exists"
        )


@router.put(
    Paths.concept_scheme,
    summary="Update a `ConceptScheme` object",
    response_model=response.ConceptScheme,
    dependencies=[Depends(verify_auth_token)],
    tags=["ConceptScheme"],
)
async def concept_scheme_update(
    request: Request,
    concept_scheme: req.ConceptScheme,
    service=Depends(get_graph_service),
    responses={404: {"description": "Resource not found"}},
) -> response.ConceptScheme:
    try:
        cs = de.ConceptScheme.from_json_ld(await request.json())
        result = await service.concept_scheme_update(cs)
        return response.ConceptScheme(**result.to_json_ld())
    except de.ConceptSchemeNotFoundError:
        raise HTTPException(status_code=404, detail=f"Concept Scheme with IRI `{cs.id_}` not found")


@router.delete(
    Paths.concept_scheme,
    summary="Delete a `ConceptScheme` object",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(verify_auth_token)],
    tags=["ConceptScheme"],
    responses={404: {"description": "Resource not found"}},
)
async def concept_scheme_delete(
    iri: str,
    service=Depends(get_graph_service),
):
    try:
        await service.concept_scheme_delete(iri=iri)
    except de.ConceptSchemeNotFoundError as err:
        raise HTTPException(status_code=404, detail=str(err))


# Relationship


@router.get(
    Paths.relationship,
    summary="Get a list of `Concept` relationships",
    response_model=list[response.Relationship],
    response_model_exclude_unset=True,
    tags=["Concept"],
)
async def relationships_get(
    iri: str,
    source: bool = True,
    target: bool = False,
    service=Depends(get_graph_service),
) -> list[response.Relationship]:
    lst = await service.relationships_get(iri=iri, source=source, target=target)
    return [response.Relationship(**obj.to_json_ld()) for obj in lst]


@router.post(
    Paths.relationship,
    summary="Create a list of `Concept` relationships",
    response_model=list[response.Relationship],
    response_model_exclude_unset=True,
    dependencies=[Depends(verify_auth_token)],
    tags=["Concept"],
    responses={409: {"description": "Resource already exists"}},
)
async def relationships_create(
    request: Request,
    relationships: list[req.Relationship],
    service=Depends(get_graph_service),
) -> list[response.Relationship]:
    try:
        incoming = de.Relationship.from_json_ld_list(await request.json())
        lst = await service.relationships_create(incoming)
        return [response.Relationship(**obj.to_json_ld()) for obj in lst]
    except de.DuplicateRelationship as err:
        raise HTTPException(status_code=409, detail=str(err))
    except (
        de.HierarchicRelationshipAcrossConceptScheme,
        de.RelationshipsReferencesConceptScheme,
    ) as err:
        raise HTTPException(status_code=422, detail=str(err))


@router.delete(
    Paths.relationship,
    summary="Delete a list of `Concept` relationships",
    dependencies=[Depends(verify_auth_token)],
    tags=["Concept"],
)
async def relationship_delete(
    request: Request,
    relationships: list[req.Relationship],
    service=Depends(get_graph_service),
) -> JSONResponse:
    incoming = de.Relationship.from_json_ld_list(await request.json())
    count = await service.relationships_delete(incoming)
    return JSONResponse(
        status_code=200,
        content={
            "detail": "Relationships (possibly) deleted",
            "count": count,
        },
    )


# Correspondence


@router.get(
    Paths.correspondence,
    summary="Get a `Correspondence` object",
    response_model=response.Correspondence,
    tags=["Correspondence"],
    responses={404: {"description": "Resource not found"}},
)
async def correspondence_get(
    iri: str,
    service=Depends(get_graph_service),
) -> response.Correspondence:
    try:
        obj = await service.correspondence_get(iri=iri)
        return response.Correspondence(**obj.to_json_ld())
    except de.CorrespondenceNotFoundError:
        raise HTTPException(status_code=404, detail=f"Correspondence with IRI `{iri}` not found")


@router.post(
    Paths.correspondence,
    summary="Create a `Correspondence` object",
    response_model=response.Correspondence,
    dependencies=[Depends(verify_auth_token)],
    tags=["Correspondence"],
    responses={409: {"description": "Resource already exists"}},
)
async def correspondence_create(
    request: Request,
    correspondence: req.Correspondence,
    service=Depends(get_graph_service),
) -> response.Correspondence:
    try:
        corr = de.Correspondence.from_json_ld(await request.json())
        result = await service.correspondence_create(corr)
        return response.Correspondence(**result.to_json_ld())
    except de.DuplicateIRI:
        raise HTTPException(
            status_code=409, detail=f"Correspondence with IRI `{corr.id_}` already exists"
        )


@router.put(
    Paths.correspondence,
    summary="Update a `Correspondence` object",
    response_model=response.Correspondence,
    dependencies=[Depends(verify_auth_token)],
    tags=["Correspondence"],
    responses={404: {"description": "Resource not found"}},
)
async def correspondence_update(
    request: Request,
    concept_scheme: req.Correspondence,
    service=Depends(get_graph_service),
) -> response.Correspondence:
    try:
        corr = de.Correspondence.from_json_ld(await request.json())
        result = await service.correspondence_update(corr)
        return response.Correspondence(**result.to_json_ld())
    except de.CorrespondenceNotFoundError:
        raise HTTPException(
            status_code=404, detail=f"Correspondence with IRI `{corr.id_}` not found"
        )


@router.delete(
    Paths.correspondence,
    summary="Delete a `Correspondence` object",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(verify_auth_token)],
    tags=["Correspondence"],
    responses={404: {"description": "Resource not found"}},
)
async def correspondence_delete(
    iri: str,
    service=Depends(get_graph_service),
):
    try:
        await service.correspondence_delete(iri=iri)
    except de.CorrespondenceNotFoundError as err:
        raise HTTPException(status_code=404, detail=str(err))


# Association


@router.get(
    Paths.association,
    summary="Get an `Association` object",
    response_model=response.Association,
    tags=["ConceptAssociation"],
    responses={404: {"description": "Resource not found"}},
)
async def association_get(
    iri: str,
    service=Depends(get_graph_service),
) -> response.Association:
    try:
        obj = await service.association_get(iri=iri)
        return response.Association(**obj.to_json_ld())
    except de.AssociationNotFoundError:
        raise HTTPException(status_code=404, detail=f"Association with IRI `{iri}` not found")


@router.post(
    Paths.association,
    summary="Create an `Association` object",
    response_model=response.Association,
    dependencies=[Depends(verify_auth_token)],
    tags=["ConceptAssociation"],
    responses={409: {"description": "Resource already exists"}},
)
async def association_create(
    request: Request,
    relationships: req.Association,
    service=Depends(get_graph_service),
) -> response.Association:
    try:
        incoming = de.Association.from_json_ld(await request.json())
        obj = await service.association_create(incoming)
        return response.Association(**obj.to_json_ld())
    except de.DuplicateIRI as exc:
        raise HTTPException(status_code=409, detail=str(exc))


@router.delete(
    Paths.association,
    summary="Delete an `Association` object",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(verify_auth_token)],
    tags=["ConceptAssociation"],
    responses={404: {"description": "Resource not found"}},
)
async def association_delete(
    iri: str,
    service=Depends(get_graph_service),
):
    try:
        await service.association_delete(iri=iri)
    except de.AssociationNotFoundError as err:
        raise HTTPException(status_code=404, detail=str(err))


@router.post(
    Paths.made_of,
    summary="Add some `Correspondence` `madeOf` links",
    response_model=response.Correspondence,
    dependencies=[Depends(verify_auth_token)],
    tags=["Correspondence"],
    responses={404: {"description": "Resource not found"}},
)
async def made_of_add(
    request: Request,
    made_of: req.MadeOf,
    service=Depends(get_graph_service),
) -> response.Correspondence:
    try:
        made_of = de.MadeOf.from_json_ld(await request.json())
        corr = await service.made_of_add(made_of)
        return response.Correspondence(**corr.to_json_ld())
    except de.CorrespondenceNotFoundError:
        raise HTTPException(
            status_code=404, detail=f"Correspondence with IRI `{made_of.id_}` not found"
        )


@router.delete(
    Paths.made_of,
    summary="Remove some `Correspondence` `madeOf` links",
    response_model=response.Correspondence,
    dependencies=[Depends(verify_auth_token)],
    tags=["Correspondence"],
    responses={404: {"description": "Resource not found"}},
)
async def made_of_remove(
    request: Request,
    made_of: req.MadeOf,
    service=Depends(get_graph_service),
) -> response.Correspondence:
    try:
        made_of = de.MadeOf.from_json_ld(await request.json())
        corr = await service.made_of_remove(made_of)
        return response.Correspondence(**corr.to_json_ld())
    except de.CorrespondenceNotFoundError:
        raise HTTPException(
            status_code=404, detail=f"Correspondence with IRI `{made_of.id_}` not found"
        )


@router.get(
    Paths.search,
    summary="Search for `Concept` objects",
    response_model=list[de.SearchResult],
    tags=["Concept"],
    responses={503: {"description": "Search engine not available"}},
)
async def concept_search(
    query: str,
    language: str,
    semantic: bool = True,
    service=Depends(get_search_service),
) -> list[de.SearchResult]:
    try:
        results = await service.search(query=query, language=language, semantic=semantic)
        return results
    except de.SearchNotConfigured:
        raise HTTPException(status_code=503, detail="Search engine not available")
    except de.UnknownLanguage:
        raise HTTPException(
            status_code=422, detail="Search engine not configured for given language"
        )


@router.get(
    Paths.suggest,
    summary="Suggestion search for `Concept` objects",
    response_model=list[de.SearchResult],
    tags=["Concept"],
    responses={503: {"description": "Search engine not available"}},
)
async def concept_suggest(
    query: str,
    language: str,
    service=Depends(get_search_service),
) -> list[de.SearchResult]:
    try:
        results = await service.suggest(query=query, language=language)
        return results
    except de.SearchNotConfigured:
        raise HTTPException(status_code=503, detail="Search engine not available")
    except de.UnknownLanguage:
        raise HTTPException(
            status_code=422, detail="Search engine not configured for given language"
        )



from enum import StrEnum
from pathlib import Path as PathLib
from urllib.parse import unquote

import structlog
from fastapi import APIRouter, Depends, HTTPException, Path, Request
from fastapi.responses import HTMLResponse, RedirectResponse
from fastapi.templating import Jinja2Templates
from langcodes import Language

from py_semantic_taxonomy.cfg import get_settings
from py_semantic_taxonomy.dependencies import get_graph_service, get_search_service
from py_semantic_taxonomy.domain import entities as de

logger = structlog.get_logger("py-semantic-taxonomy")

router = APIRouter(prefix="/web", include_in_schema=False)


def value_for_language(value: list[dict[str, str]], lang: str) -> str:
    """Get the `@value` for a list of multilingual strings with correct `@language` value"""
    for dct in value:
        if dct.get("@language") == lang:
            return dct.get("@value", "")
    return ""


templates = Jinja2Templates(directory=str(PathLib(__file__).parent / "templates"))
templates.env.filters["split"] = lambda s, sep: s.split(sep)
templates.env.filters["lang"] = value_for_language


def format_languages(languages: list[str]) -> list[tuple[str, str]]:
    """Take a list of ISO 639 language codes and return (code, name)"""
    return [(code, Language.get(code).display_name(code).title()) for code in languages]


class WebPaths(StrEnum):
    concept_schemes = "/concept_schemes/"
    concept_scheme_view = "/concept_scheme/{iri:path}"
    concept_view = "/concept/{iri:path}"
    search = "/search/"


@router.get('/')
async def redirect_blank_web_page(
    request: Request,
) -> RedirectResponse:
    return RedirectResponse(request.url_for('web_concept_schemes'))


@router.get(
    WebPaths.concept_schemes,
    response_class=HTMLResponse,
)
async def web_concept_schemes(
    request: Request,
    service=Depends(get_graph_service),
    settings=Depends(get_settings),
) -> HTMLResponse:
    """List all concept schemes."""
    concept_schemes = await service.concept_scheme_list()
    return templates.TemplateResponse(
        "concept_schemes.html",
        {
            "request": request,
            "concept_schemes": concept_schemes,
            "languages": format_languages(settings.languages),
        },
    )


@router.get(
    WebPaths.concept_scheme_view,
    response_class=HTMLResponse,
)
async def web_concept_scheme_view(
    request: Request,
    iri: str = Path(..., description="The IRI of the concept scheme"),
    service=Depends(get_graph_service),
    settings=Depends(get_settings),
) -> HTMLResponse:
    """View a specific concept scheme."""
    try:
        decoded_iri = unquote(iri)
        concept_scheme = await service.concept_scheme_get(iri=decoded_iri)
        concepts = await service.concepts_get_for_scheme(
            concept_scheme_iri=decoded_iri, top_concepts_only=True
        )
        return templates.TemplateResponse(
            "concept_scheme_view.html",
            {
                "request": request,
                "concept_scheme": concept_scheme,
                "concepts": concepts,
                "languages": format_languages(settings.languages),
            },
        )
    except de.ConceptSchemeNotFoundError:
        raise HTTPException(status_code=404, detail=f"Concept Scheme with IRI `{iri}` not found")
    except de.ConceptSchemesNotInDatabase as e:
        logger.error("Database error while fetching concept scheme", iri=iri, error=str(e))
        raise HTTPException(status_code=500, detail="Database error while fetching concept scheme")


@router.get(
    WebPaths.concept_view,
    response_class=HTMLResponse,
)
async def web_concept_view(
    request: Request,
    iri: str = Path(..., description="The IRI of the concept to view"),
    service=Depends(get_graph_service),
    settings=Depends(get_settings),
) -> HTMLResponse:
    """View a specific concept."""
    try:
        decoded_iri = unquote(iri)
        concept = await service.concept_get(iri=decoded_iri)
        relationships = await service.relationships_get(iri=decoded_iri, source=True, target=True)

        relationships_data = []
        for rel in relationships:
            relationships_data.append(
                {"source": rel.source, "target": rel.target, "predicate": rel.predicate}
            )

        related_concepts = {}
        for rel in relationships:
            related_iri = rel.target if rel.source == decoded_iri else rel.source
            if related_iri not in related_concepts:
                try:
                    related_concept = await service.concept_get(iri=related_iri)
                    related_concepts[related_iri] = related_concept.to_json_ld()
                except de.ConceptNotFoundError:
                    logger.warning(
                        "Related concept not found",
                        concept_iri=iri,
                        related_iri=related_iri,
                    )

        return templates.TemplateResponse(
            "concept_view.html",
            {
                "request": request,
                "concept": concept,
                "relationships": relationships_data,
                "related_concepts": related_concepts,
                "languages": format_languages(settings.languages),
            },
        )
    except de.ConceptNotFoundError:
        raise HTTPException(status_code=404, detail=f"Concept with IRI `{iri}` not found")
    except de.ConceptSchemesNotInDatabase as e:
        logger.error("Database error while fetching concept", iri=decoded_iri, error=str(e))
        raise HTTPException(status_code=500, detail="Database error while fetching concept")


@router.get(
    WebPaths.search,
    response_class=HTMLResponse,
)
async def web_search(
    request: Request,
    query: str = "",
    language: str = "en",
    semantic: bool = True,
    search_service=Depends(get_search_service),
) -> HTMLResponse:
    """Search for concepts."""
    try:
        results = []
        if query:
            results = await search_service.search(
                query=query, language=language, semantic=semantic
            )
        return templates.TemplateResponse(
            "search.html",
            {
                "request": request,
                "query": query,
                "language": language,
                "semantic": semantic,
                "results": results,
            },
        )
    except de.SearchNotConfigured:
        raise HTTPException(status_code=503, detail="Search engine not available")
    except de.UnknownLanguage:
        raise HTTPException(
            status_code=422, detail="Search engine not configured for given language"
        )



from py_semantic_taxonomy.dependencies import get_kos_graph, get_search_service
from py_semantic_taxonomy.domain.constants import (
    SKOS_HIERARCHICAL_RELATIONSHIP_PREDICATES,
    RelationshipVerbs,
)
from py_semantic_taxonomy.domain.entities import (
    Association,
    AssociationNotFoundError,
    Concept,
    ConceptNotFoundError,
    ConceptScheme,
    ConceptSchemeNotFoundError,
    ConceptSchemesNotInDatabase,
    Correspondence,
    CorrespondenceNotFoundError,
    DuplicateRelationship,
    GraphObject,
    HierarchicRelationshipAcrossConceptScheme,
    HierarchyConflict,
    MadeOf,
    Relationship,
    RelationshipsInCurrentConceptScheme,
    RelationshipsReferencesConceptScheme,
)
from py_semantic_taxonomy.domain.ports import KOSGraphDatabase, SearchService


class GraphService:
    def __init__(
        self,
        graph: KOSGraphDatabase | None = None,
        search: SearchService | None = None,
    ):
        self.graph = graph or get_kos_graph()
        self.search = search or get_search_service()

    async def get_object_type(self, iri: str) -> GraphObject:
        return await self.graph.get_object_type(iri=iri)

    # Concept

    async def concept_get(self, iri: str) -> Concept:
        return await self.graph.concept_get(iri=iri)

    async def _concept_refers_to_concept_scheme_in_database(self, concept: Concept) -> None:
        concept_schemes = set(await self.concept_scheme_get_all_iris())
        given_cs = {cs["@id"] for cs in concept.schemes}
        if not given_cs.intersection(concept_schemes):
            raise ConceptSchemesNotInDatabase(
                f"At least one of the specified concept schemes must be in the database: {given_cs}"
            )

    async def _check_top_concept(self, concept: Concept) -> None:
        if rels := (await self.relationships_get(concept.id_, verb=RelationshipVerbs.broader)):
            raise HierarchyConflict(
                f"Concept is marked as `topConceptOf` but also has broader relationship to `{rels[0].target}`"
            )

    async def concept_create(
        self, concept: Concept, relationships: list[Relationship] = []
    ) -> Concept:
        await self._concept_refers_to_concept_scheme_in_database(concept)

        if concept.top_concept_of:
            await self._check_top_concept(concept)

            for rel in relationships:
                if rel.source == concept.id_ and rel.predicate == RelationshipVerbs.broader:
                    raise HierarchyConflict(
                        f"Concept is marked as `topConceptOf` but also has broader relationship to `{rel.target}`"
                    )

        await self.graph.concept_create(concept=concept)
        if relationships:
            try:
                await self.relationships_create(relationships)
            except (HierarchicRelationshipAcrossConceptScheme, DuplicateRelationship) as err:
                await self.concept_delete(concept.id_)
                raise err

        if self.search.is_configured():
            await self.search.create_concept(concept)

        return concept

    async def concept_update(self, concept: Concept) -> Concept:
        await self._concept_refers_to_concept_scheme_in_database(concept)

        if concept.top_concept_of:
            await self._check_top_concept(concept)

        current = await self.graph.concept_get(concept.id_)
        current_schemes = {cs["@id"] for cs in current.schemes}
        new_schemes = {cs["@id"] for cs in concept.schemes}
        if current_schemes.difference(new_schemes):
            # Can remove a ConceptScheme only if it doesn't create cross-scheme hierarchical
            # relationships
            rel_schemes = (
                await self.graph.known_concept_schemes_for_concept_hierarchical_relationships(
                    concept.id_
                )
            )
            if missing := set(rel_schemes).difference(new_schemes):
                raise RelationshipsInCurrentConceptScheme(
                    f"Update asked to change concept schemes, but existing concept scheme {missing} had hierarchical relationships."
                )
        concept = await self.graph.concept_update(concept=concept)

        if self.search.is_configured():
            await self.search.update_concept(concept)

        return concept

    async def concept_delete(self, iri: str) -> None:
        rowcount = await self.graph.concept_delete(iri=iri)
        if not rowcount:
            raise ConceptNotFoundError(f"Concept with IRI `{iri}` not found")

        if self.search.is_configured():
            await self.search.delete_concept(iri)

        return

    async def concepts_get_for_scheme(
        self, concept_scheme_iri: str, top_concepts_only: bool = False
    ) -> list[Concept]:
        """Get all concepts that belong to a given concept scheme."""
        return await self.graph.concepts_get_for_scheme(
            concept_scheme_iri=concept_scheme_iri,
            top_concepts_only=top_concepts_only,
        )

    # Concept Scheme

    async def concept_scheme_get(self, iri: str) -> ConceptScheme:
        return await self.graph.concept_scheme_get(iri=iri)

    async def concept_scheme_get_all_iris(self) -> list[str]:
        return await self.graph.concept_scheme_get_all_iris()

    async def concept_scheme_list(self) -> list[ConceptScheme]:
        return await self.graph.concept_scheme_list()

    async def concept_scheme_create(self, concept_scheme: ConceptScheme) -> ConceptScheme:
        return await self.graph.concept_scheme_create(concept_scheme=concept_scheme)

    async def concept_scheme_update(self, concept_scheme: ConceptScheme) -> ConceptScheme:
        return await self.graph.concept_scheme_update(concept_scheme=concept_scheme)

    async def concept_scheme_delete(self, iri: str) -> None:
        rowcount = await self.graph.concept_scheme_delete(iri=iri)
        if not rowcount:
            raise ConceptSchemeNotFoundError(f"Concept Scheme with IRI `{iri}` not found")
        return

    # Relationships

    async def relationships_get(
        self,
        iri: str,
        source: bool = True,
        target: bool = False,
        verb: RelationshipVerbs | None = None,
    ) -> list[Relationship]:
        return await self.graph.relationships_get(iri=iri, source=source, target=target, verb=verb)

    async def _relationships_check_source_target_share_known_concept_scheme(
        self, relationships: list[Relationship]
    ) -> None:
        for rel in relationships:
            if rel.predicate in SKOS_HIERARCHICAL_RELATIONSHIP_PREDICATES and not (
                await self.graph.relationship_source_target_share_known_concept_scheme(rel)
            ):
                raise HierarchicRelationshipAcrossConceptScheme(
                    f"Hierarchical relationship between `{rel.source}` and `{rel.target}` crosses Concept Schemes. Use an associative relationship like `skos:broadMatch` instead."
                )

    async def relationships_create(self, relationships: list[Relationship]) -> list[Relationship]:
        concept_schemes = await self.concept_scheme_get_all_iris()
        for rel in relationships:
            if rel.source in concept_schemes:
                raise RelationshipsReferencesConceptScheme(
                    f"Relationship `{rel}` source refers to concept scheme `{rel.source}`"
                )
            if rel.target in concept_schemes:
                raise RelationshipsReferencesConceptScheme(
                    f"Relationship `{rel}` target refers to concept scheme `{rel.target}`"
                )

        await self._relationships_check_source_target_share_known_concept_scheme(relationships)
        return await self.graph.relationships_create(relationships)

    async def relationships_delete(self, relationships: list[Relationship]) -> int:
        return await self.graph.relationships_delete(relationships)

    # Correspondence

    async def correspondence_get(self, iri: str) -> Correspondence:
        return await self.graph.correspondence_get(iri=iri)

    async def correspondence_create(self, correspondence: Correspondence) -> Correspondence:
        return await self.graph.correspondence_create(correspondence=correspondence)

    async def correspondence_update(self, correspondence: Correspondence) -> Correspondence:
        return await self.graph.correspondence_update(correspondence=correspondence)

    async def correspondence_delete(self, iri: str) -> None:
        rowcount = await self.graph.correspondence_delete(iri=iri)
        if not rowcount:
            raise CorrespondenceNotFoundError(f"Correspondence with IRI `{iri}` not found")
        return

    async def made_of_add(self, made_of: MadeOf) -> Correspondence:
        return await self.graph.made_of_add(made_of)

    async def made_of_remove(self, made_of: MadeOf) -> Correspondence:
        return await self.graph.made_of_remove(made_of)

    # Association

    async def association_get(self, iri: str) -> Association:
        return await self.graph.association_get(iri=iri)

    async def association_create(self, association: Association) -> Association:
        return await self.graph.association_create(association=association)

    async def association_delete(self, iri: str) -> None:
        rowcount = await self.graph.association_delete(iri=iri)
        if not rowcount:
            raise AssociationNotFoundError(f"Association with IRI `{iri}` not found")
        return



from copy import copy
from dataclasses import asdict, dataclass, field, fields
from datetime import datetime
from urllib.parse import quote_plus, unquote

from py_semantic_taxonomy.domain.constants import (
    RDF_MAPPING,
    SKOS_RELATIONSHIP_PREDICATES,
    AssociationKind,
    RelationshipVerbs,
)
from py_semantic_taxonomy.domain.hash_utils import hash_fnv64


def select_language(objs: list[dict], language: str, concatenate: bool = False) -> str:
    strings = [
        obj["@value"] for obj in objs if obj["@language"].lower().startswith(language.lower())
    ]
    if concatenate:
        return " ".join(strings)
    return strings


# Allow mixing non-default and default values in dataclasses
# See https://www.trueblade.com/blogs/news/python-3-10-new-dataclass-features
@dataclass(kw_only=True)
class Serializable:
    def to_db_dict(self) -> dict:
        return asdict(self)

    def to_json_ld(self, fields_: list[str] = [], extra: bool = True) -> dict:
        """Return this data formatted as (but not serialized to) SKOS expanded JSON LD"""
        class_fields = fields_ or {f.name for f in fields(self)}.difference({"extra"})
        dct = copy(self.extra) if extra else {}
        for attr, label in RDF_MAPPING.items():
            if attr not in class_fields:
                continue
            if value := getattr(self, attr):
                dct[label] = value

        return dct

    @classmethod
    def from_json_ld(cls, dict_: dict, fields_: list[str] = [], extra: bool = True) -> "SKOS":
        source_dict, data = copy(dict_), {}
        class_fields = fields_ or {f.name for f in fields(cls)}.difference({"extra"})
        for dataclass_label, skos_label in RDF_MAPPING.items():
            if dataclass_label in class_fields and skos_label in source_dict:
                data[dataclass_label] = source_dict.pop(skos_label)
        if extra:
            data["extra"] = {
                key: value
                for key, value in source_dict.items()
                if key not in SKOS_RELATIONSHIP_PREDICATES
            }
        return cls(**data)


@dataclass(kw_only=True)
class SKOS(Serializable):
    id_: str
    types: list[str]
    pref_labels: list[dict[str, str]]
    status: list[dict]
    notations: list[dict[str, str]] = field(default_factory=list)
    definitions: list[dict[str, str]] = field(default_factory=list)
    change_notes: list[dict] = field(default_factory=list)
    history_notes: list[dict] = field(default_factory=list)
    editorial_notes: list[dict] = field(default_factory=list)
    extra: dict = field(default_factory=dict)


@dataclass(kw_only=True)
class Concept(SKOS):
    schemes: list[dict]
    alt_labels: list[dict[str, str]] = field(default_factory=list)
    hidden_labels: list[dict[str, str]] = field(default_factory=list)
    top_concept_of: list[dict] = field(default_factory=list)

    def to_search_dict(self, language: str) -> dict:
        return {
            # Can't use URL as id, even if escaped:
            # https://github.com/typesense/typesense/issues/192
            "id": hash_fnv64(self.id_),
            "url": quote_plus(self.id_),
            "alt_labels": select_language(self.alt_labels, language),
            "hidden_labels": select_language(self.hidden_labels, language),
            # One per language but can have language variants
            "pref_label": select_language(self.pref_labels, language, concatenate=True),
            "definition": select_language(self.definitions, language, concatenate=True),
            # Not language-specific
            "notation": " ".join([obj["@value"] for obj in self.notations]),
            "all_languages_pref_labels": [obj["@value"] for obj in self.pref_labels],
        }


@dataclass(kw_only=True)
class ConceptScheme(SKOS):
    created: list[datetime]
    creators: list[dict]
    version: list[str]


@dataclass(frozen=True)
class Relationship:
    source: str
    target: str
    predicate: RelationshipVerbs

    def to_db_dict(self) -> dict:
        return asdict(self)

    def to_json_ld(self) -> dict:
        """Return this data formatted as (but not serialized to) SKOS expanded JSON LD"""
        return {"@id": self.source, str(self.predicate): [{"@id": self.target}]}

    @classmethod
    def from_json_ld(cls, obj: dict) -> list["Relationship"]:
        def _get_object_id(obj: dict) -> str:
            try:
                return obj["@id"]
            except KeyError:
                raise ValueError(f"Can't find `@id` in given JSON-LD object: {obj}")

        result = set()
        mapping = {elem.value: elem for elem in RelationshipVerbs}
        source = _get_object_id(obj)

        for key, value in obj.items():
            if key == RelationshipVerbs.narrower:
                for elem in value:
                    result.add((_get_object_id(elem), RelationshipVerbs.broader, source))
            elif key in mapping:
                for elem in value:
                    result.add((source, mapping[key], _get_object_id(elem)))

        return sorted(
            [Relationship(source=s, target=o, predicate=p) for s, p, o in result],
            key=lambda x: (x.source, x.target, x.predicate),
        )

    @classmethod
    def from_json_ld_list(cls, obj: list[dict]) -> list["Relationship"]:
        return sorted(
            {rel for elem in obj for rel in cls.from_json_ld(elem)},
            key=lambda x: (x.source, x.target, x.predicate),
        )


@dataclass(kw_only=True)
class Correspondence(ConceptScheme):
    compares: list[dict]
    made_ofs: list[dict] = field(default_factory=list)


@dataclass(kw_only=True)
class MadeOf(Serializable):
    id_: str
    made_ofs: list[dict]

    def to_json_ld(self) -> dict:
        return super().to_json_ld(extra=False)

    @classmethod
    def from_json_ld(cls, dict_: dict) -> "Association":
        return super().from_json_ld(dict_=dict_, extra=False)


@dataclass(kw_only=True)
class Association(Serializable):
    id_: str
    types: list[str]
    source_concepts: list[dict]
    target_concepts: list[dict]
    kind: AssociationKind = AssociationKind.simple
    extra: dict = field(default_factory=dict)

    def __post_init__(self):
        self.kind = (
            AssociationKind.conditional if len(self.source_concepts) > 1 else AssociationKind.simple
        )

    def to_json_ld(self) -> dict:
        # Exclude `extra`
        return super().to_json_ld(fields_={"id_", "types", "source_concepts", "target_concepts"})

    @classmethod
    def from_json_ld(cls, dict_: dict) -> "Association":
        # Exclude `extra`
        return super().from_json_ld(
            dict_=dict_, fields_={"id_", "types", "source_concepts", "target_concepts"}
        )


# For type hinting
GraphObject = Concept | ConceptScheme | Correspondence | Association


@dataclass
class SearchResult:
    id_: str
    label: str
    highlight: str | None = None

    @staticmethod
    def from_typesense_results(results: dict) -> list["SearchResult"]:
        return [
            SearchResult(
                id_=unquote(res["document"]["url"]),
                label=res["document"]["pref_label"],
                highlight=(
                    res["highlight"]["pref_label"]["snippet"]
                    if "pref_label" in res["highlight"]
                    else None
                ),
            )
            for res in results["hits"]
        ]

    def to_json(self) -> dict:
        return asdict(self)


class NotFoundError(Exception):
    pass


class ConceptNotFoundError(NotFoundError):
    pass


class ConceptSchemeNotFoundError(NotFoundError):
    pass


class RelationshipNotFoundError(NotFoundError):
    pass


class CorrespondenceNotFoundError(NotFoundError):
    pass


class AssociationNotFoundError(NotFoundError):
    pass


class DuplicateIRI(Exception):
    pass


class DuplicateRelationship(Exception):
    pass


class HierarchicRelationshipAcrossConceptScheme(Exception):
    pass


class RelationshipsInCurrentConceptScheme(Exception):
    pass


class RelationshipsReferencesConceptScheme(Exception):
    pass


class ConceptSchemesNotInDatabase(Exception):
    pass


class HierarchyConflict(Exception):
    pass


class UnknownLanguage(Exception):
    pass


class SearchNotConfigured(Exception):
    pass



from typing import Self

from pydantic import BaseModel, ConfigDict, Field, conlist, field_validator, model_validator

from py_semantic_taxonomy.adapters.routers.doc_examples import (
    CHANGE_NOTE,
    DEFINITION,
    EDITORIAL_NOTE,
    HISTORY_NOTE,
)
from py_semantic_taxonomy.adapters.routers.validation import (
    IRI,
    DateTime,
    MultilingualString,
    Node,
    NonLiteralNote,
    Notation,
    Status,
    VersionString,
    one_per_language,
)
from py_semantic_taxonomy.domain.constants import RDF_MAPPING as RDF
from py_semantic_taxonomy.domain.constants import (
    SKOS,
    SKOS_RELATIONSHIP_PREDICATES,
    XKOS,
)
from py_semantic_taxonomy.domain.constants import RelationshipVerbs as RV


class KOSCommon(BaseModel):
    id_: IRI = Field(
        alias=RDF["id_"],
        title="Object IRI (`@id`)",
        description="https://www.w3.org/TR/json-ld/#node-identifiers",
        example="http://data.europa.eu/xsp/cn2024/010021000090",
    )
    types: conlist(item_type=IRI) = Field(
        alias=RDF["types"],
        title="Object `@type`",
        description="https://www.w3.org/TR/json-ld/#specifying-the-type",
        example=["http://www.w3.org/2004/02/skos/core#Concept"],
    )
    pref_labels: conlist(MultilingualString, min_length=1) = Field(
        alias=RDF["pref_labels"],
        title="SKOS preferred labels (one per language)",
        description="https://www.w3.org/TR/skos-primer/#secpref",
        example=[
            {"@value": "CHAPTER 1 - LIVE ANIMALS", "@language": "en"},
            {"@value": "CAP\u00cdTULO 1 - ANIMAIS VIVOS", "@language": "pt"},
        ],
    )
    status: conlist(Status, min_length=1) = Field(
        alias=RDF["status"],
        title="BIBO status (accepted/draft/rejected)",
        description="https://github.com/dcmi/bibo/blob/main/rdf/bibo.ttl#L391",
        example=[{"@id": "http://purl.org/ontology/bibo/status/accepted"}],
    )
    notations: list[Notation] = Field(
        alias=RDF["notations"],
        default=[],
        description="https://www.w3.org/TR/skos-primer/#secnotations",
        title="SKOS notation (typed literal)",
        example=[
            {"@value": "01", "@type": "http://www.w3.org/1999/02/22-rdf-syntax-ns#PlainLiteral"}
        ],
    )
    change_notes: list[NonLiteralNote] = Field(
        alias=RDF["change_notes"],
        default=[],
        definition="https://www.w3.org/TR/skos-primer/#secdocumentation",
        title="SKOS change note with additional required fields",
        example=CHANGE_NOTE,
    )
    history_notes: list[NonLiteralNote] = Field(
        alias=RDF["history_notes"],
        default=[],
        definition="https://www.w3.org/TR/skos-primer/#secdocumentation",
        title="SKOS history note with additional required fields",
        example=HISTORY_NOTE,
    )
    editorial_notes: list[NonLiteralNote] = Field(
        alias=RDF["editorial_notes"],
        default=[],
        definition="https://www.w3.org/TR/skos-primer/#secdocumentation",
        title="SKOS editorial note with additional required fields",
        example=EDITORIAL_NOTE,
    )

    model_config = ConfigDict(extra="allow")

    def model_dump(self, exclude_unset=True, by_alias=True, *args, **kwargs):
        return super().model_dump(*args, exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)

    @field_validator("pref_labels", mode="after")
    @classmethod
    def pref_labels_one_per_language(
        cls, value: list[MultilingualString]
    ) -> list[MultilingualString]:
        return one_per_language(value, "prefLabel")


class Concept(KOSCommon):
    """Validation class for SKOS Concepts.

    Checks that required fields are included and have correct type."""

    schemes: conlist(Node, min_length=1) = Field(
        alias=RDF["schemes"],
        title="SKOS concept scheme (normally only one)",
        description="https://www.w3.org/TR/skos-primer/#secscheme",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/cn2024"}],
    )
    # Can have multiple alternative labels per language, and multiple languages
    alt_labels: list[MultilingualString] = Field(
        alias=RDF["alt_labels"],
        default=[],
        title="SKOS alternative labels. Can be more than one per language.",
        description="https://www.w3.org/TR/skos-primer/#secalt",
        example=[
            {"@value": "Horsies, moo-moos, etc.", "@language": "en"},
        ],
    )
    # Can have multiple hidden labels per language, and multiple languages
    hidden_labels: list[MultilingualString] = Field(
        alias=RDF["hidden_labels"],
        default=[],
        title="SKOS hidden labels",
        description="https://www.w3.org/TR/skos-primer/#sechidden",
        example=[{"@language": "ine-pro", "@value": "ékwos"}],
    )
    # One definition per language, at least one definition
    definitions: list[MultilingualString] = Field(
        alias=RDF["definitions"],
        default=[],
        title="SKOS definition (one per language)",
        description="https://www.w3.org/TR/skos-primer/#secdocumentation",
        example=DEFINITION,
    )
    top_concept_of: conlist(Node, max_length=1) = Field(
        alias=RDF["top_concept_of"],
        default=[],
        title="SKOS concept scheme if this concept is at top of hierarchy (maximum 1)",
        description="https://www.w3.org/TR/skos-primer/#secscheme",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/cn2024"}],
    )

    @field_validator("types", mode="after")
    @classmethod
    def type_includes_concept(cls, value: list[str]) -> list[str]:
        CONCEPT = f"{SKOS}Concept"
        if CONCEPT not in value:
            raise ValueError(f"`@type` values must include `{CONCEPT}`")
        return value

    @field_validator("definitions", mode="after")
    @classmethod
    def definition_one_per_language(
        cls, value: list[MultilingualString]
    ) -> list[MultilingualString]:
        return one_per_language(value, "definition")

    @model_validator(mode="after")
    def notations_disjoint_pref_label(self) -> Self:
        if overlap := {obj.value for obj in self.pref_labels}.intersection(
            {obj.value for obj in self.notations}
        ):
            raise ValueError(f"Found overlapping values in `prefLabel` and `notation`: {overlap}")
        return self


class ConceptCreate(Concept):
    broader: list[Node] = Field(
        alias=str(RV.broader),
        default=[],
        title="SKOS broader",
        description="https://www.w3.org/TR/skos-primer/#sechierarchy",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/010021000090"}],
    )
    narrower: list[Node] = Field(
        alias=str(RV.narrower),
        default=[],
        title="SKOS narrower (use of `broader` is preferred; see docs)",
        description="https://www.w3.org/TR/skos-primer/#sechierarchy",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/010100000080"}],
    )

    @model_validator(mode="after")
    def hierarchy_doesnt_reference_self(self) -> Self:
        for node in self.broader:
            if node.id_ == self.id_:
                raise ValueError("Concept can't have `broader` relationship to itself")
        for node in self.narrower:
            if node.id_ == self.id_:
                raise ValueError("Concept can't have `narrower` relationship to itself")
        return self

    @model_validator(mode="before")
    @classmethod
    def check_no_transitive_relationships(cls, data: dict) -> dict:
        for key in data:
            if key in {f"{SKOS}broaderTransitive", f"{SKOS}narrowerTransitive"}:
                short = key[len(SKOS) :]
                raise ValueError(
                    f"Found `{short}` in new concept; transitive relationships are implied and should be omitted."
                )
        return data


class ConceptUpdate(Concept):
    @model_validator(mode="before")
    @classmethod
    def check_no_graph_relationships(cls, data: dict) -> dict:
        for key in data:
            if key in SKOS_RELATIONSHIP_PREDICATES:
                short = key[len(SKOS) :]
                raise ValueError(
                    f"Found `{short}` in concept update; Use specific API calls to update graph structure."
                )
        return data


class ConceptSchemeCommon(KOSCommon):
    created: conlist(DateTime, min_length=1, max_length=1) = Field(
        alias="http://purl.org/dc/terms/created",
        title="DCTerms created timestamp",
        description="https://www.dublincore.org/specifications/dublin-core/dcmi-terms/#http://purl.org/dc/terms/created",
        example=[
            {"@type": "http://www.w3.org/2001/XMLSchema#dateTime", "@value": "2023-10-11T13:59:56"}
        ],
    )
    creators: list[Node] = Field(
        alias="http://purl.org/dc/terms/creator",
        title="DCTerms creators list",
        description="https://www.dublincore.org/specifications/dublin-core/dcmi-terms/#http://purl.org/dc/elements/1.1/creator",
        example=[{"@id": "http://publications.europa.eu/resource/authority/corporate-body/ESTAT"}],
    )
    version: conlist(VersionString, min_length=1, max_length=1) = Field(
        alias="http://www.w3.org/2002/07/owl#versionInfo",
        title="OWL version info",
        description="https://www.w3.org/TR/owl-ref/#versionInfo-def",
        example=[{"@value": "2024"}],
    )


class ConceptScheme(ConceptSchemeCommon):
    """Validation class for SKOS Concept Schemes.

    Checks that required fields are included and have correct type."""

    definitions: conlist(MultilingualString, min_length=1) = Field(
        alias=RDF["definitions"],
        title="SKOS definition (one per language)",
        description="https://www.w3.org/TR/skos-primer/#secdocumentation",
        example=DEFINITION,
    )

    @field_validator("types", mode="after")
    @classmethod
    def type_includes_concept_scheme(cls, value: list[str]) -> list[str]:
        SCHEME = f"{SKOS}ConceptScheme"
        if SCHEME not in value:
            raise ValueError(f"`@type` must include `{SCHEME}`")
        return value

    @model_validator(mode="before")
    @classmethod
    def check_no_top_concept(cls, data: dict) -> dict:
        """skos:hasTopConcept has range skos:Concept, which we don't want. Create links later."""
        if f"{SKOS}hasTopConcept" in data:
            raise ValueError(
                f"Found `hasTopConcept` in concept scheme; Specify `topConceptOf` of constituent concepts instead."
            )
        return data

    @field_validator("definitions", mode="after")
    @classmethod
    def definition_one_per_language(
        cls, value: list[MultilingualString]
    ) -> list[MultilingualString]:
        return one_per_language(value, "definition")


class Relationship(BaseModel):
    id_: IRI = Field(
        alias=RDF["id_"],
        title="Object IRI (`@id`)",
        description="https://www.w3.org/TR/json-ld/#node-identifiers",
        example="http://data.europa.eu/xsp/cn2024/010021000090",
    )
    broader: list[Node] = Field(
        alias=str(RV.broader),
        default=[],
        title="SKOS broader",
        description="https://www.w3.org/TR/skos-primer/#sechierarchy",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/010021000090"}],
    )
    narrower: list[Node] = Field(
        alias=str(RV.narrower),
        default=[],
        title="SKOS narrower (use of `broader` is preferred; see docs)",
        description="https://www.w3.org/TR/skos-primer/#sechierarchy",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/010100000080"}],
    )
    exact_match: list[Node] = Field(
        alias=str(RV.exact_match),
        default=[],
        title="SKOS exact match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010100000080"}],
    )
    close_match: list[Node] = Field(
        alias=str(RV.close_match),
        default=[],
        title="SKOS close match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010100000080"}],
    )
    broad_match: list[Node] = Field(
        alias=str(RV.broad_match),
        default=[],
        title="SKOS broad match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010021000090"}],
    )
    narrow_match: list[Node] = Field(
        alias=str(RV.narrow_match),
        default=[],
        title="SKOS narrow match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010100000080"}],
    )
    related_match: list[Node] = Field(
        alias=str(RV.related_match),
        default=[],
        title="SKOS related match",
        description="https://www.w3.org/TR/skos-primer/#secassociative",
        example=[{"@id": "https://www.wikidata.org/wiki/Q726"}],
    )

    _RELATIONSHIP_FIELDS = (
        "broader",
        "narrower",
        "exact_match",
        "close_match",
        "broad_match",
        "narrow_match",
        "related_match",
    )

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    def model_dump(self, exclude_unset=True, by_alias=True, *args, **kwargs):
        return super().model_dump(*args, exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)

    @model_validator(mode="after")
    def no_self_references(self) -> Self:
        for field in self._RELATIONSHIP_FIELDS:
            for obj in getattr(self, field, []):
                if obj.id_ == self.id_:
                    raise ValueError("Relationship has same source and target")

        return self

    @model_validator(mode="after")
    def two_relationships_of_same_type(self) -> Self:
        for field in self._RELATIONSHIP_FIELDS:
            if len(getattr(self, field, [])) > 1:
                raise ValueError(f"Found multiple relationships of type `{field}`")
        return self

    @model_validator(mode="after")
    def exactly_one_relationship_type(self) -> Self:
        truthy = sorted({field for field in self._RELATIONSHIP_FIELDS if getattr(self, field)})

        if len(truthy) > 1:
            raise ValueError(f"Found multiple relationships {truthy} where only one is allowed")
        elif not truthy:
            raise ValueError(f"Found zero relationships")

        return self


class Correspondence(ConceptSchemeCommon):
    definitions: list[MultilingualString] = Field(
        alias=RDF["definitions"],
        default=[],
        title="SKOS definition (one per language)",
        description="https://www.w3.org/TR/skos-primer/#secdocumentation",
        example=DEFINITION,
    )
    compares: conlist(Node, min_length=1) = Field(
        alias=f"{XKOS}compares",
        title="List of `ConceptScheme` objects being compared",
        description="https://rdf-vocabulary.ddialliance.org/xkos.html#correspondences",
        example=[
            {"@id": "http://data.europa.eu/xsp/cn2023/cn2023"},
            {"@id": "http://data.europa.eu/xsp/cn2024/cn2024"},
        ],
    )

    @field_validator("types", mode="after")
    @classmethod
    def type_includes_correspondence(cls, value: list[str]) -> list[str]:
        CONCEPT = f"{XKOS}Correspondence"
        if CONCEPT not in value:
            raise ValueError(f"`@type` values must include `{CONCEPT}`")
        return value

    @field_validator("definitions", mode="after")
    @classmethod
    def definition_one_per_language(
        cls, value: list[MultilingualString]
    ) -> list[MultilingualString]:
        return one_per_language(value, "definition")

    @model_validator(mode="before")
    @classmethod
    def check_no_made_of(cls, data: dict) -> dict:
        for key in data:
            if key == RDF["made_ofs"]:
                raise ValueError(
                    f"Found `{RDF['made_ofs']}` in new correspondence; use dedicated API calls for this data."
                )
        return data


class MadeOf(BaseModel):
    id_: IRI = Field(
        alias=RDF["id_"],
        title="Object IRI (`@id`)",
        description="https://www.w3.org/TR/json-ld/#node-identifiers",
        example="http://data.europa.eu/xsp/cn2024/010021000090",
    )
    made_ofs: list[Node] = Field(
        alias=RDF["made_ofs"],
        title="List of `ConceptAssociation` objects in a `Correspondence`",
        description="https://rdf-vocabulary.ddialliance.org/xkos.html#correspondences",
        example="http://data.europa.eu/xsp/cn2024/010021000090",
    )

    model_config = ConfigDict(extra="forbid")

    def model_dump(self, by_alias=True, *args, **kwargs):
        return super().model_dump(*args, by_alias=by_alias, **kwargs)


class Association(BaseModel):
    id_: IRI = Field(
        alias=RDF["id_"],
        title="Object IRI (`@id`)",
        description="https://www.w3.org/TR/json-ld/#node-identifiers",
        example="http://data.europa.eu/xsp/cn2023_cn2024/something",
    )
    types: conlist(item_type=IRI) = Field(
        alias=RDF["types"],
        title="Object `@type`",
        description="https://www.w3.org/TR/json-ld/#specifying-the-type",
        example=["http://rdf-vocabulary.ddialliance.org/xkos#ConceptAssociation"],
    )
    source_concepts: conlist(Node, min_length=1) = Field(
        alias=RDF["source_concepts"],
        title="List of source `Concept` objects",
        description="https://rdf-vocabulary.ddialliance.org/xkos.html#correspondences",
        example=[{"@id": "http://data.europa.eu/xsp/cn2023/010011000090"}],
    )
    target_concepts: conlist(Node, min_length=1) = Field(
        alias=RDF["target_concepts"],
        title="List of target `Concept` objects",
        description="https://rdf-vocabulary.ddialliance.org/xkos.html#correspondences",
        example=[{"@id": "http://data.europa.eu/xsp/cn2024/010011000090"}],
    )

    model_config = ConfigDict(extra="allow")

    def model_dump(self, exclude_unset=True, by_alias=True, *args, **kwargs):
        return super().model_dump(*args, exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)

    @field_validator("types", mode="after")
    @classmethod
    def type_includes_association(cls, value: list[str]) -> list[str]:
        SCHEME = f"{XKOS}ConceptAssociation"
        if SCHEME not in value:
            raise ValueError(f"`@type` must include `{SCHEME}`")
        return value



from pathlib import Path

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles

from py_semantic_taxonomy.adapters.persistence.database import (
    create_engine,
    init_db,
)
from py_semantic_taxonomy.adapters.routers.router import router
from py_semantic_taxonomy.adapters.routers.web_router import router as web_router
from py_semantic_taxonomy.adapters.routers.catch_router import router as catch_router
from py_semantic_taxonomy.dependencies import get_search_service

# from fastapi.middleware.cors import CORSMiddleware


def create_app() -> FastAPI:
    app = FastAPI()

    @app.on_event("startup")
    async def database():
        await init_db(create_engine())

    @app.on_event("startup")
    async def search():
        ts = get_search_service()
        if ts.configured:
            await ts.initialize()

    # app.add_middleware(
    #     CORSMiddleware,
    #     allow_origins=settings.allow_origins,
    #     allow_credentials=True,
    #     allow_methods=["*"],
    #     allow_headers=["*"],
    # )

    app.include_router(router)
    app.include_router(web_router)
    app.include_router(catch_router)
    app.mount(
        "/static",
        StaticFiles(directory=Path(__file__).parent / "adapters" / "routers" / "static"),
        name="static",
    )
    return app


def test_app() -> FastAPI:
    app = FastAPI()
    app.include_router(router)
    app.include_router(web_router)
    return app


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(
        "py_semantic_taxonomy.app:create_app",
        host="0.0.0.0",
        port=8000,
        log_level="info",
        reload=True,
    )



from collections import defaultdict
from datetime import datetime
from enum import Enum, StrEnum
from typing import Annotated, Literal

import langcodes
import rfc3987
from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    Field,
    PlainSerializer,
    conlist,
    field_validator,
)


def validate_iri(value: str) -> str:
    rfc3987.parse(value, rule="IRI")
    return value


IRI = Annotated[
    str,
    AfterValidator(validate_iri),
]
SerializableDateTime = Annotated[
    datetime,
    PlainSerializer(lambda dt: dt.isoformat(), return_type=str),
]


class VersionString(BaseModel):
    """
    The range for http://www.w3.org/2002/07/owl#versionInfo is
    http://www.w3.org/2000/01/rdf-schema#Resource, which is a class instance with no restrictions
    or other guidance: https://www.w3.org/TR/rdf-schema/#ch_resource.

    In JSON-LD version strings would therefore be given as:

    ```'http://www.w3.org/2002/07/owl#versionInfo': [{'@value': 'some_number'}]```

    * It's a list because there can in theory be more than one
    * It uses ```{}``` syntax because it is an object, not a string

    Version numbers should be numeric, so we don't need a language or text direction. However, we
    don't know how they are arranged, so can't convert to numeric types.
    """

    value: str = Field(alias="@value")


class DateTimeType(Enum):
    date = "http://www.w3.org/2001/XMLSchema#date"
    datetime = "http://www.w3.org/2001/XMLSchema#dateTime"


class DateTime(BaseModel):
    """
    The range for http://purl.org/dc/terms/created is http://www.w3.org/2000/01/rdf-schema#Literal.
    We will be more strict and require an ISO 8601 date or datetime.
    """

    type_: DateTimeType = Field(alias="@type")
    # Pydantic can convert ISO 8601 dates and datetimes to native types automatically
    value: SerializableDateTime = Field(alias="@value")

    model_config = ConfigDict(use_enum_values=True)


class MultilingualString(BaseModel):
    """
    A dictionary which provides a language-tagged string literal compatible with JSON-LD 1.1.

    The language code must be compliant with BCP 47:
    https://www.w3.org/TR/rdf11-concepts/#dfn-language-tagged-string

    We use the `langcodes` library to ensure this compliance, and to fix common errors:
    https://pypi.org/project/langcodes/

    The direction is not required by the JSON-LD standard, but we will always include it to make
    code more predictable.
    """

    value: str = Field(alias="@value")
    language: str = Field(alias="@language")
    # https://www.w3.org/TR/json-ld11/#base-direction
    direction: Literal["ltr", "rtl"] = Field(alias="@direction", default="ltr")

    @field_validator("language", mode="after")
    @classmethod
    def language_code_valid(cls, value):
        # Will raise a subclass of ValidationError if invalid
        return langcodes.standardize_tag(value)

    model_config = ConfigDict(extra="forbid")


class Node(BaseModel):
    """
    If the RDF refers to a class instance instead of a string literal, this instance can have
    arbitrary attributes, but it must have at least an `@id`.
    """

    id_: IRI = Field(alias="@id")

    model_config = ConfigDict(extra="allow")


class StatusChoice(StrEnum):
    accepted = "http://purl.org/ontology/bibo/status/accepted"
    draft = "http://purl.org/ontology/bibo/status/draft"
    rejected = "http://purl.org/ontology/bibo/status/rejected"


class Status(BaseModel):
    id_: StatusChoice = Field(alias="@id")

    model_config = ConfigDict(extra="allow", allow_enum_values=True)


def one_per_language(values: list[MultilingualString], label: str) -> list[MultilingualString]:
    """Ensure that there is only one multilingual string per language code"""
    grouped = defaultdict(list)
    for mlstr in values:
        grouped[mlstr.language].append(mlstr.value)
    for key, obj in grouped.items():
        if len(obj) > 1:
            raise ValueError(
                f"Only one string per language code is allowed for `{label}`, but language `{key}` has {len(obj)} strings: {obj}"
            )
    return values


class Notation(BaseModel):
    """
    TBD: The range for http://www.w3.org/2004/02/skos/core#notation
    """

    value: str = Field(alias="@value")
    type_: IRI = Field(
        alias="@type", default="http://www.w3.org/1999/02/22-rdf-syntax-ns#PlainLiteral"
    )

    model_config = ConfigDict(extra="forbid")


class NonLiteralNote(BaseModel):
    """
    Used for skos:changeNote, skos:historyNote, and skos:editorialNote.

    SKOS allows for string literals but we require an author and timestamp for each.
    """

    values: conlist(MultilingualString, min_length=1) = Field(
        alias="http://www.w3.org/1999/02/22-rdf-syntax-ns#value"
    )
    creators: conlist(Node, min_length=1) = Field(alias="http://purl.org/dc/terms/creator")
    # XKOS guidance recommends `issued` instead of `created` for notes
    issued: conlist(DateTime, min_length=1, max_length=1) = Field(
        alias="http://purl.org/dc/terms/issued"
    )
    value: conlist(MultilingualString, min_length=1, max_length=1) = Field(
        alias="http://www.w3.org/1999/02/22-rdf-syntax-ns#value"
    )
